#!/usr/bin/python3
#
# dsbench/dsbench.py - Main program
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

import os
import traceback

if os.name == "posix":
    import signal
    try:
        # Not available on wheezy!
        import faulthandler
        faulthandler.register(signal.SIGUSR1)
    except:
        pass

from tpcbCore import *
from tpcbNode import *
from tpcbPg import *
from tpcbFdb import *
from tpcbFirebirdSql import *
# from tpcbOra import *


# i18n preparations.
def _(s):
    return s


try:
    # Configuration of test environment
    env = tpcbEnv()
    env.fileconfig()
    env.argvconfig()

    if not env.daemon and not env.killer and env.archive == NOTHING:
        if not env.guiconfig():
            print(_("Cancel immediately."))
            sys.exit(0)

    # Start the logging module
    env.start()

except Exception as e:
    # Logger not started so far, print message to console
    print(_("Could not configure application! Terminate immediately."))
    traceback.print_exc()
    sys.exit(-1)


try:
    retval = 0

    if env.archive != NOTHING:
        # Replot data:
        # Pickup data from archive
        env.readarchive()

        # Environment for TPC-B data acquisition
        a = tpcbMultiple(env)
        # Read measurements from files
        a.read()

        # Save results including diagram creation with gnuplot
        a.replotrates()
        a.replotconfidences()

        # Archive updated results
        env.makearchive()

    elif env.daemon:
        # NODE:
        try:
            node = tpcbNodeLauncher(env)
            node.start()
        except KeyboardInterrupt:
            # KeyboardInterrupts abort the loop.
            env.logger.warning(_("Application interrupted manually!"))

    elif env.killer:
        # Kill all nodes:
        killer = tpcbNodeKiller(env)
        killer.broadcast()

    else:
        # MASTER:
        # Environment for TPC-B data acquisition
        a = tpcbMultiple(env)

        # Create database environment
        if env.job.db == "PG":
            db = tpcbDbPostgres(env, a)
        elif env.job.db == "FB" and env.job.fbdriver == "fdb":
            db = tpcbDbFdb(env, a)
        elif env.job.db == "FB":
            db = tpcbDbFirebirdSql(env, a)
        # elif env.job.db == "ORA":
        #    db = tpcbDbOracle(env, a)

        # Collect database parameters
        # (versions, configuration parameters)
        db.parameters()

        # Initialisation of test environment after database selection
        env.initialise()

        # Measurement loop for all scaling factors
        # (only if not marked as finished by loading results from files)
        # - create database
        # - measure TPC-B transaction rate
        # - drop database
        # - compute and log results
        #
        for m in a.singles:
            if not m.finished:
                try:
                    db.create(m, True)
                    db.do()
                    db.drop(True)
                    db.loglast()
                except KeyboardInterrupt:
                    # KeyboardInterrupts abort the loop.
                    # Data collected so far will be saved.
                    env.logger.warning(_("Application interrupted manually! "
                                         "Try to save data gathered so far."))
                    break
                except:
                    env.logger.exception(
                        _("Application encountered unexpected situation! "
                          "Try to save results collected so far."))
                    break

        # Data sheet containing test environment and database parameters
        env.datasheet()
        # Results representation
        a.presentresults()
        # Archive results
        env.makearchive()

except:
    # Log crash information
    env.logger.exception(_("Application encountered unresolvable situation! "
                           "Terminate immediately."))
    retval = -1

finally:
    # Stop the logging module
    env.stop()
    # Instead of sys.exit() we use os._exit() which also kills
    # dead locked threads in case of errorneous (timeout) abort.
    os._exit(retval)
