# -*- coding: utf-8 -*-
#
# dsbench/tpcbCore.py - Core
#
# Copyright (C) 2014-2016 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

import os
import re
import sys
import math
import time
import socket
import psutil
import shutil
import threading
import argparse
import tempfile
import platform
import logging
import random

from tpcbDefs import *
from tpcbBuf import *
# from tpcbI18n import *

try:
    from tpcbGui import *
except:
    pass

# Database related imports
PSYCOPG2 = True
try:
    import psycopg2
except:
    PSYCOPG2 = False

FDB = True
try:
    import fdb
except:
    FDB = False

FIREBIRDSQL = True
try:
    import firebirdsql
except:
    FIREBIRDSQL = False


# Initialisations
random.seed()


# Version
__version__ = "1.1.10"


# i18n preparations
def _(s):
    return s


class tpcbNodeInfo():
    """Node information.

    This class acquires information about the test systems which can be
    the Master responsible to control the test procedure or a remote node
    which connects to the test database to carry out transactions.
    """
    def __init__(self, **kwargs):
        """Initialisation.

        This constructor assigns defaults for system parameters and replaces
        them by user defined values that are given as keyword arguments. A
        string created with the __str__() method of this class can be passed
        to this constructor to create a new copy.

        Args:
            **kwargs: Keyword arguments that can be named as each attribute
                of this class.
        """
        self.ip = ""
        self.host = ""
        self.cpu = 0
        self.core = 0
        self.name = ""
        self.platform = ""
        self.arch = ""
        self.machine = ""
        self.processor = ""
        self.python = ""
        self.psycopg2 = NOTINSTALLED
        self.fdb = NOTINSTALLED
        self.firebirdsql = NOTINSTALLED
        self.ram = 0.0
        self.swap = 0.0

        # Replace default attributes by kwargs
        for key, value in kwargs.items():
            if not hasattr(self, key):
                raise Exception(
                    _("Attribute '{0:s}' not available.").format(key))
            setattr(self, key, value)

    def getsystem(self):
        """Collect system parameters."""
        self.name = os.name

        # Host name without domain name
        # self.host = socket.gethostname()
        self.host = platform.node()
        if "." in self.host:
            if not self.host.startswith("."):
                self.host, *dummy = self.host.split(".")

        self.platform = platform.platform()
        self.arch = str(platform.architecture())
        self.machine = platform.machine()
        self.processor = platform.processor()
        info = sys.version_info
        self.python = "{0:d}.{1:d}.{2:d}".format(
            info.major, info.minor, info.micro)
        self.ram = psutil.virtual_memory().total / GB
        self.swap = psutil.swap_memory().total / GB
        if sys.version_info >= (3, 4):
            self.cpu = psutil.cpu_count(logical=False)
            self.core = psutil.cpu_count()
        if PSYCOPG2:
            self.psycopg2 = psycopg2.__version__
        if FDB:
            self.fdb = fdb.__version__
        if FIREBIRDSQL:
            self.firebirdsql = firebirdsql.__version__

    def __str__(self):
        """Convert to parsible string.

        This method delivers a string that can be parsed by Python eval()
        expression to create a copy of this class. This is used by network
        transferring of system parameters.
        """
        out = "tpcbNodeInfo(ip='" + self.ip + "'"
        out += ", host='" + self.host + "'"
        out += ", cpu=" + str(self.cpu)
        out += ", core=" + str(self.core)
        out += ", name='" + self.name + "'"
        out += ", platform='" + self.platform + "'"
        out += ", arch=\"" + self.arch + "\""
        out += ", machine='" + self.machine + "'"
        out += ", processor='" + self.processor + "'"
        out += ", python='" + self.python + "'"
        out += ", psycopg2='" + self.psycopg2 + "'"
        out += ", fdb='" + self.fdb + "'"
        out += ", firebirdsql='" + self.firebirdsql + "'"
        out += ", ram=" + str(self.ram)
        out += ", swap=" + str(self.swap)
        out += ")"
        return out


class tpcbJobInfo():
    """Job information.

    This class combines all information required to send a transaction job
    to a remote node.
    """
    def __init__(self, **kwargs):
        """Initialisation.

        This constructor defines defaults for all job parameters and
        replaces them by user defined values that are given as keyword
        arguments. A string created with the __str__() method of this class
        can be passed to this constructor to create a new copy.

        Args:
            **kwargs: Keyword arguments that can be named as each attribute
                of this class.
        """
        # Database selection
        self.db = "PG"
        # General database configuration
        self.pgServer = "localhost"
        self.pgPassword = "postgres"
        self.fbServer = "localhost"
        self.fbPassword = "masterkey"
        self.fbConfig = "/etc/firebird/2.5/firebird.conf"
        self.fbdriver = "firebirdsql"
        # Number of nodes (0 = process locally)
        self.nodes = 0
        # Number of threads per node; 0 = doing by main thread
        self.threads = 0
        # Number of clients per thread
        self.clients = 1
        # Duration of one meaurement cycle in seconds
        self.duration = 60.0
        # Type of transaction for testing
        self.transtype = TPCB_DEFAULT
        # Current scale factor (for remote processing)
        self.scale = 0
        # Current cycle number (for remote processing)
        self.cycle = 0
        # Current database size (for remote processing)
        self.dbsize = 0.0
        # Use prepare for transactions
        self.prepare = False
        # Open/close connection to database per transaction
        self.reconnect = False
        # Retry failed transactions until success
        self.retry = False

        # Replace default attributes by kwargs
        for key, value in kwargs.items():
            if not hasattr(self, key):
                raise Exception(
                    _("Attribute '{0:s}' not available.").format(key))
            setattr(self, key, value)

    @property
    def timeout(self):
        """Calculate timeout.

        Determines a reasonable timeout depending on the configuration.
        """
        return TIMEFACTOR * self.duration + TIMEOUT

    def __str__(self):
        """Convert to parsible string.

        This method delivers a string that can be parsed by Python eval()
        expression to create a copy of this class. This is used by network
        transferring of jobs.
        """
        out = "tpcbJobInfo(db='" + self.db + "'"
        out += ", pgServer='" + self.pgServer + "'"
        out += ", pgPassword='" + self.pgPassword + "'"
        out += ", fbServer='" + self.fbServer + "'"
        out += ", fbPassword='" + self.fbPassword + "'"
        out += ", fbConfig='" + self.fbConfig + "'"
        out += ", fbdriver='" + self.fbdriver + "'"
        out += ", nodes=" + str(self.nodes)
        out += ", threads=" + str(self.threads)
        out += ", clients=" + str(self.clients)
        out += ", duration=" + str(self.duration)
        out += ", transtype=" + tpcbTransTable[self.transtype].symbol
        out += ", scale=" + str(self.scale)
        out += ", cycle=" + str(self.cycle)
        out += ", dbsize=" + str(self.dbsize)
        out += ", prepare=" + str(self.prepare)
        out += ", reconnect=" + str(self.reconnect)
        out += ", retry=" + str(self.retry)
        out += ")"
        return out


class tpcbRandom():
    """Random number generator.

    This random number generator delivers input parameters of a TPC-B
    transaction with regard to non recurring branch_id values as good
    as possible. This strategy avoids that different transaction threads
    using the same random number generator instance access the same rows
    in the database that would raise transaction errors.
    """
    def __init__(self, optimized, scale):
        """Initialisation.

        Initialises the random number generator.

        Args:
            optimized: Uses recurring branch_id values to reduce transaction
                errors.
            scale: Scaling factor of the database which corresponds to the
                total number of rows in the branch table.
        """
        self.lock = threading.Lock()
        self.optimized = optimized
        self.scale = scale
        self.counter = 0
        self.branch_init = []
        for n in range(self.scale):
            self.branch_init.append(n)
        self.branch_id = list(self.branch_init)

    def next(self):
        """Get next set of random numbers.

        Delivers the next set of random numbers for a TPC-B transaction.
        This method is thread safe by using critical sections (lock).

        Returns:
            A tuple of input parameters for a single TPC-B transaction:
            branch_id, teller_id, account_id, delta.
        """
        self.lock.acquire()

        try:
            if (self.optimized):
                # Optimized version:
                # The order of all available branch_id's (same as scale)
                # is selected randomly. Each branch_id is returned once
                # in order to avoid access to the same row in the branch
                # table.
                if self.scale <= 1:
                    # scale == 1: only one branch_id - no other choice.
                    branch_id = 0
                elif self.scale == 2:
                    # scale == 2: branch_id is returned alternating.
                    self.counter = (self.counter + 1) % self.scale
                    branch_id = self.counter
                else:
                    # scale > 2:
                    # The order of all available branch_id's is selected
                    # randomly.
                    if (self.counter == 0):
                        # Memorise last returned branch_id.
                        last_branch_id = self.branch_id[-1]
                        # Create a new cycle.
                        self.branch_id = []
                        pool = list(self.branch_init)
                        for n in reversed(range(self.scale)):
                            k = random.randint(0, n)
                            self.branch_id.append(pool[k])
                            del pool[k]
                        # Avoids that the last returned branch_id from last
                        # cycle is the same as the first of the new cycle.
                        if self.branch_id[0] == last_branch_id:
                            del self.branch_id[0]
                            self.branch_id.append(last_branch_id)
                        # Return the first branch_id of the new cycle.
                        branch_id = self.branch_id[0]
                        self.counter += 1
                    else:
                        branch_id = self.branch_id[self.counter]
                        self.counter = (self.counter + 1) % self.scale
                teller_id = 10 * branch_id + random.randint(0, 9)
            else:
                # Classic version:
                # teller_id randomly, branch_id dependent
                teller_id = random.randint(0, 10 * self.scale - 1)
                branch_id = teller_id // 10

            account_id = random.randint(0, 100000 - 1)

            if self.scale > 1:
                if random.randint(1, 100) > 85:
                    while (True):
                        other_branch = random.randint(0, self.scale - 1)
                        if other_branch != branch_id:
                            break
                    account_id += other_branch * 100000
                else:
                    account_id += branch_id * 100000

            delta = random.randint(-999999, 999999)
        finally:
            self.lock.release()

        return branch_id, teller_id, account_id, delta


class tpcbLoad():
    """System load acquisition.

    This class helps to acquire system's IO and CPU load. These values dont
    have sense if the database server is not locally.
    """
    def __init__(self, env):
        """Initialisation.

        Args:
            env: Need to know the test environment for logging and
                the drive hosting the database.
        """
        self.env = env

    def iostat(self):
        """Read IO status.

        Requests the IO load parameters for the given drive
        (self.env.dbdev).
        """
        ios = psutil.disk_io_counters(perdisk=True)
        for key, value in ios.items():
            if self.env.dbdev.find(key) != -1:
                return value
        return None

    def start(self):
        """Read system load at start.

        This is the first request to collect system load information
        before a transaction measurement cycle starts.
        """
        self.env.logger.debug(_("Reading system load parameters (start)."))
        self.iostat()
        time.sleep(1)
        self.io_start = self.iostat()
        self.cpu_start = psutil.cpu_times()
        psutil.cpu_percent(interval=None)

    def stop(self):
        """Reading system load at stop.

        This is the second request to collect system load information
        after a transaction measurement cycle finished. Results are compared
        to results at start.
        """
        self.env.logger.debug(_("Reading system load parameters (stop)."))
        self.cpu = psutil.cpu_percent(interval=None)
        self.io_end = self.iostat()
        self.cpu_end = psutil.cpu_times()

        if self.io_start and self.io_end:
            self.read_count = (self.io_end.read_count -
                               self.io_start.read_count)
            self.write_count = (self.io_end.write_count -
                                self.io_start.write_count)
            self.read_bytes = (self.io_end.read_bytes -
                               self.io_start.read_bytes) / GB
            self.write_bytes = (self.io_end.write_bytes -
                                self.io_start.write_bytes) / GB
            self.read_time = (self.io_end.read_time -
                              self.io_start.read_time) / 1000.0
            self.write_time = (self.io_end.write_time -
                               self.io_start.write_time) / 1000.0

        if self.cpu_start and self.cpu_end:
            if hasattr(self.cpu_start, "user"):
                self.user = self.cpu_end.user - self.cpu_start.user
            if hasattr(self.cpu_start, "system"):
                self.system = self.cpu_end.system - self.cpu_start.system
            if hasattr(self.cpu_start, "idle"):
                self.idle = self.cpu_end.idle - self.cpu_start.idle
            if hasattr(self.cpu_start, "nice"):
                self.nice = self.cpu_end.nice - self.cpu_start.nice
            if hasattr(self.cpu_start, "iowait"):
                self.iowait = self.cpu_end.iowait - self.cpu_start.iowait
            if hasattr(self.cpu_start, "irq"):
                self.irq = self.cpu_end.irq - self.cpu_start.irq
            if hasattr(self.cpu_start, "softirq"):
                self.softirq = self.cpu_end.softirq - self.cpu_start.softirq
            if hasattr(self.cpu_start, "steal"):
                self.steal = self.cpu_end.steal - self.cpu_start.steal
            if hasattr(self.cpu_start, "guest"):
                self.guest = self.cpu_end.guest - self.cpu_start.guest
            if hasattr(self.cpu_start, "guest_nice"):
                self.guest_nice = (self.cpu_end.guest_nice -
                                   self.cpu_start.guest_nice)

    def comma(self):
        """Helper: Appends a comma."""
        if self.first:
            self.first = False
            return ""
        else:
            return ","

    def __str__(self):
        """Print results.

        Returns differences of all system load parameters between second and
        first request (between after and before transaction measurement cycle
        took place).
        """
        out = ""
        if not self.env.islocal:
            out += "\n"
            out += _("Database system is installed remotely.\n"
                     "CPU- and IO-Loads do not represent the database server!")
        if self.io_start and self.io_end:
            out += "\n"
            out += "[IO] read_count={0:d}, write_count={1:d}, " \
                   "read_bytes={2:.3f}GB, write_bytes={3:.3f}GB, " \
                   "read_time={4:.2f}s, write_time={5:.2f}s".format(
                       self.read_count, self.write_count,
                       self.read_bytes, self.write_bytes,
                       self.read_time, self.write_time)

        if self.cpu_start and self.cpu_end:
            self.first = True
            out += "\n[CPU]"
            if hasattr(self.cpu_start, "user"):
                out += self.comma() + " user={0:.2f}s".format(self.user)
            if hasattr(self.cpu_start, "nice"):
                out += self.comma() + " nice={0:.2f}s".format(self.nice)
            if hasattr(self.cpu_start, "system"):
                out += self.comma() + " system={0:.2f}s".format(self.system)
            if hasattr(self.cpu_start, "idle"):
                out += self.comma() + " idle={0:.2f}s".format(self.idle)
            if hasattr(self.cpu_start, "iowait"):
                out += self.comma() + " iowait={0:.2f}s".format(self.iowait)
            if hasattr(self.cpu_start, "irq"):
                out += self.comma() + " irq={0:.2f}s".format(self.irq)
            if hasattr(self.cpu_start, "softirq"):
                out += self.comma() + " softirq={0:.2f}s".format(self.softirq)
            if hasattr(self.cpu_start, "steal"):
                out += self.comma() + " steal={0:.2f}s".format(self.steal)
            if hasattr(self.cpu_start, "guest"):
                out += self.comma() + " guest={0:.2f}s".format(self.guest)
            if hasattr(self.cpu_start, "guest_nice"):
                out += self.comma() + " guest_nice={0:.2f}s".format(
                    self.guest_nice)

        if self.cpu:
            out += "\n"
            out += "[CPU] cpu_load={0:.2f}%".format(self.cpu)

        # On Windows OS the IO load request can not assigned to
        # a definute drive letter because the IO load data structure
        # uses partition names (like PhysicalDrive0) as keys instead of
        # drive letters.
        if len(out) > 3:
            return out
        else:
            return _("unavailable")


class tpcbMsg():
    """Network messages.

    This is a helper class for creating and parsing messages required
    to communicate with remote nodes to instruct them with transaction
    jobs.
    """
    def __init__(self, **kwargs):
        """Initialisation.

        Creates a network message using an input string and parameters or
        a byte sequence as keyword arguments.

        Args:
            **kwargs:
                msgtype - Message type.
                parameters - Content depending on the message type.
                buffer - Byte sequence.
        """
        self.command = None
        self._parameters = None
        self.encoding = "utf-8"
        self.msgtypes = ("SCAN", "KILL", "NODE", "JOB", "RESULT", "ENDMARK")
        self.asBytes = bytes("", self.encoding)
        self.asStr = ""

        for key, value in kwargs.items():
            if key == "buffer":
                if isinstance(value, str):
                    self.asStr = value
                else:
                    self.asBytes = value
            elif key == "msgtype":
                if value in self.msgtypes:
                    self.command = value
                else:
                    raise Exception(_("Unknown message type!"))
            elif key == "parameters":
                self._parameters = value

    def append(self, buffer):
        """Append content.

        Append strings or bytes to the message. Applied if network
        transfer is fragmented due to high message lengths.

        Args:
            buffer: Content to append.
        """
        if isinstance(buffer, str):
            self.asStr += buffer
        else:
            self.asBytes += buffer

    def resolve(self):
        """Analyse message.

        The message is parsed and transcoded for network transfer or
        as strings.
        """
        try:
            if self.asStr:
                self.asBytes = bytes(self.asStr, self.encoding)
                if " " in self.asStr:
                    self.command, self._parameters = self.asStr.split(" ", 1)
                else:
                    self.command = self.asStr
            elif self.asBytes:
                self.asStr = self.asBytes.decode(self.encoding).strip()
                if " " in self.asStr:
                    self.command, self._parameters = self.asStr.split(" ", 1)
                else:
                    self.command = self.asStr
            elif self.command:
                if self._parameters:
                    self.asStr = "{0:s} {1:s}".format(
                        self.command, self._parameters)
                else:
                    self.asStr = self.command
                self.asBytes = bytes(self.asStr, self.encoding)

            # Sometimes byte sequences containing residence times can
            # be converted to a valid string but do not contain a valid
            # command. In this case the except block is not executed
            # but it should be.
            if self.command not in self.msgtypes:
                self.command = None
                self._parameters = self.asBytes

        except:
            self.command = None
            self._parameters = self.asBytes

    @property
    def isscan(self):
        """Checks if the message is a UDP broadcast."""
        self.resolve()
        if self.command == self.msgtypes[0]:
            return True
        else:
            return False

    @property
    def iskill(self):
        """Checks if the message is a killing instruction."""
        self.resolve()
        if self.command == self.msgtypes[1]:
            return True
        else:
            return False

    @property
    def isnode(self):
        """Checks if the message is a node identifier."""
        self.resolve()
        if self.command == self.msgtypes[2]:
            return True
        else:
            return False

    @property
    def isjob(self):
        """Checks if the message is a transaction job."""
        self.resolve()
        if self.command == self.msgtypes[3]:
            return True
        else:
            return False

    @property
    def isresult(self):
        """Checks if the message is a result."""
        self.resolve()
        if self.command == self.msgtypes[4]:
            return True
        else:
            return False

    @property
    def isendmark(self):
        """Checks if the message is an endmark."""
        self.resolve()
        if self.command == self.msgtypes[5]:
            return True
        else:
            return False

    @property
    def parameters(self):
        """Delivers the parameters of the message as string."""
        self.resolve()
        return self._parameters

    def __str__(self):
        """Delivers a string representation of the message."""
        self.resolve()
        return self.asStr

    def __bytes__(self):
        """Delivers a byte sequence represenation of the message.

        It is suitable for network transfer.
        """
        self.resolve()
        return self.asBytes


class tpcbEnv():
    """Test environment for database benchmarks.

    This class serves as environment for database benchmarks. It contains
    parameters of master and remote systems, configuration, test conditions
    and database specifications.
    """
    def __init__(self):
        """Initialisation

        Collects system parameters and defines defaults for configuration
        and test conditions.
        """
        self.logging = False
        self.lock = threading.Lock()
        self.deferred = []
        # List of database parameters
        # Filled by _parameter() method of database classes.
        self.dbenv = []
        # Database file or path and the file system on which it is located
        # Parameters are set by the database module.
        self.dbdev = ""
        self.dbpath = ""
        self.fstype = UNKNOWN
        # Format used to create archive file containing results
        self.archiveformat = "zip"
        # Number of trials to resume remote transactions
        # (Normally only one trial is enough but sometimes UDP package
        #  loss may require an other trial.)
        self.ntrials = 3

        # Node information:
        # Collect information of myself.
        self.own = tpcbNodeInfo()
        self.own.getsystem()
        # Information buffer of remote nodes that will be filled
        # during tests.
        self.nodeinfo = []

        # Job information:
        # Instanciate with defaults
        self.job = tpcbJobInfo()

        # Configuration defaults:
        # Appending background debug logs
        self.guard = False
        # Also save raw residence time data
        self.saveraw = False
        # Use accurate data for residence time diagram plotting
        self.accurate = False
        # Operate as node
        self.daemon = False
        # Operate as mode killer
        self.killer = False
        # Run test with parameters to determine system DSI
        self.dsi = False
        # Suppress graphical user interface
        self.suppress = False
        # Logging Level
        self.logdebug = False
        # Path to gnuplot executable (for Windows OS)
        self.gnuplot = "C:\\Apps\\gnuplot\\bin\\wgnuplot.exe"
        # File name of archive for replotting data
        self.archive = NOTHING
        # Remote node control port (UDP and TCP)
        self.port = 24336

        # Test condition defaults:
        # File name for result output
        self.output = "results"
        # Number of measurement cycles per scale factor
        self.no = 10
        # List of scale factors
        self.scales = (1, 2, 5, 10, 20, 50, 100)
        # Remove database after measurements
        self.remove = True
        # List of confidence levels for statistics
        self.confidences = (68.0, 95.0, 99.0)
        # Concurrent time measurements excluded from diagram
        self.exclctimes = ("tclean", "tdrop")
        # Maximum transaction rate for plotting
        self.tpcbmax = 0

    @property
    def islocal(self):
        """Checks, if the database is locally."""
        if self.job.db == "PG":
            if self.job.pgServer in LOCALTUPLE:
                return True
        elif self.job.db == "FB":
            if self.job.fbServer in LOCALTUPLE:
                return True
        # elif self.job.db == "ORA":
        #    if self.job.oraServer in LOCALTUPLE:
        #        return True
        return False

    @property
    def dbhost(self):
        """Delivers the database host name or IP."""
        if self.job.db == "PG":
            if self.job.pgServer not in LOCALTUPLE:
                return self.job.pgServer
        elif self.job.db == "FB":
            if self.job.fbServer not in LOCALTUPLE:
                return self.job.fbServer
        # elif self.job.db == "ORA":
        #    if self.job.oraServer not in LOCALTUPLE:
        #        return self.job.oraServer
        return self.own.host

    def __repr__(self):
        """Representation of the test environment."""
        out = "tpcbEnv(logging=" + str(self.logging)
        out += ", dbenv=" + str(self.dbenv)
        out += ", dbdev='" + self.dbdev + "'"
        out += ", dbpath='" + self.dbpath + "'"
        out += ", fstype='" + self.fstype + "'"
        out += ", archiveformat='" + self.archiveformat + "'"
        out += ", ntrials=" + str(self.ntrials)
        out += ", own=" + str(self.own)
        out += ", nodeinfo=" + str(self.nodeinfo)
        out += ", job=" + str(self.job)
        out += ", guard=" + str(self.guard)
        out += ", saveraw=" + str(self.saveraw)
        out += ", accurate=" + str(self.accurate)
        out += ", daemon=" + str(self.daemon)
        out += ", killer=" + str(self.killer)
        out += ", dsi=" + str(self.dsi)
        out += ", suppress=" + str(self.suppress)
        out += ", logdebug=" + str(self.logdebug)
        out += ", gnuplot='" + self.gnuplot + "'"
        out += ", archive='" + self.archive + "'"
        out += ", port=" + str(self.port)
        out += ", output='" + self.output + "'"
        out += ", no=" + str(self.no)
        out += ", scales=" + str(self.scales)
        out += ", remove=" + str(self.remove)
        out += ", confidences=" + str(self.confidences)
        out += ", exclctimes=" + str(self.exclctimes)
        out += ", tpcbmax=" + str(self.tpcbmax)
        out += ")"
        return out

    def outpath(self, ext):
        """Absolute path for writing.

        Results are written in well names files in a temporary folder.

        Args:
            ext: File extension.

        Returns:
            The full path of a results file for writing including
            the given extension.
        """
        # self.basedir = os.path.join(os.path.expanduser("~"), "." + NAME)
        path = os.path.join(self.tempdir.name, self.output + "." + ext)
        if self.logging:
            self.logger.debug(_("Create temporary file '%s'."), path)
        return path

    def inpath(self, ext):
        """Absolute path for reading.

        For replotting of results some files need to be read from a
        temporary folder.

        Args:
            ext: File extension.

        Returns:
            The full path of a results file for reading including
            the given extension.
        """
        path = os.path.join(self.tempdir.name, self.output + "." + ext)
        if self.logging:
            self.logger.debug(_("Read temporary file '%s'."), path)
        return path

    def start(self):
        """Starts the environment.

        Creates temporary folder if configured and stars the logging.
        Do this after complete configuration of the environment because
        logging level and file name for logging can be changed.
        """
        # Create logger with name 'NAME'
        self.logger = logging.getLogger(NAME)
        self.logger.setLevel(logging.DEBUG)

        # File handler for master to create test log
        if not self.daemon and not self.killer:
            # Create temporary directory
            self.tempdir = tempfile.TemporaryDirectory()
            # Create file handler
            logfile = self.outpath("log")
            if os.path.exists(logfile):
                os.remove(logfile)
            fh = logging.FileHandler(logfile)
            if self.logdebug:
                fh.setLevel(logging.DEBUG)
            else:
                fh.setLevel(logging.INFO)
            self.logger.addHandler(fh)

        # File handler for additional background debug logging
        # that can help to catch errors if using dsbench in batch
        # processing
        if self.guard:
            gf = logging.FileHandler(GUARDFILE)
            gf.setLevel(logging.DEBUG)
            self.logger.addHandler(gf)

        # Create console handler
        ch = logging.StreamHandler()
        if self.logdebug:
            ch.setLevel(logging.DEBUG)
        else:
            ch.setLevel(logging.INFO)
        self.logger.addHandler(ch)

        # Create formatters
        self.blank = logging.Formatter("%(message)s")
        self.formatter = logging.Formatter(
            "%(asctime)s -- [%(levelname)s] (%(threadName)-10s): "
            "%(message)s")

        # Configuring all handlers
        for h in self.logger.handlers:
            h.setFormatter(self.formatter)

        # Send first line to logger
        self.logger.info(
            _("%s version '%s' opened."), NAME, __version__)

        # Print deferred logs
        for line in self.deferred:
            self.logger.info(line)
        self.logging = True

    def stop(self):
        """Stop the environment.

        Stops logging and removes all temporary files.
        """
        self.logger.info(
            _("%s version '%s' closed."), NAME, __version__)
        self.logging = False
        # Closing handlers required for releasing log file
        for h in self.logger.handlers:
            h.flush()
            h.close()
            self.logger.removeHandler(h)
        del self.logger
        if not self.daemon and not self.killer:
            del self.tempdir

    def blankinfo(self, msg):
        """Log without preamble.

        Prints a line without default additional information like time,
        log level and thread name by temporary changing the formatter
        setup.

        Args:
            msg: Message to log.
        """
        for h in self.logger.handlers:
            h.setFormatter(self.blank)
        self.logger.info("\n" + msg)
        for h in self.logger.handlers:
            h.setFormatter(self.formatter)

    def fileconfig(self):
        """Read configuration.

        Reads the configuration from files depending on the system.
        First, system wide configuration is tried to read. Next, user
        specific configuration is looked for. And last the configuration
        file in the working directory is executed. Each configuration file
        overrides configuration which was read earlier.
        """
        local = "tpcb.conf"
        note = _("Found and read configuration file: ")

        # system wide configuration
        if os.name == "posix":
            paths = ("/etc/tpcb", "/etc")
            for path in paths:
                file = os.path.join(path, local)
                if os.path.exists(file):
                    self.deferred.append(note + file)
                    exec(open(file).read())
        # user specific configuration
        if os.name == "posix":
            home = os.path.expanduser("~")
            file = os.path.join(home, local)
            if os.path.exists(file):
                self.deferred.append(note + file)
                exec(open(file).read())
        # configuration in the working directory
        if os.path.exists(local):
            self.deferred.append(note + local)
            exec(open(local).read())

    def argvconfig(self):
        """Command line configuration.

        Reads the test conditions from the command line. The command line
        syntax is inspired by "pgbench". Command line configuration overrides
        configuration by files.
        """
        parser = argparse.ArgumentParser(
            description=_("%s: Database Benchmark Utility.").format(NAME))
        parser.add_argument("-f", "--config", default="",
                            help=_("Configuration file. If a file is given "
                                   "all other command line options will be "
                                   "ignored."))
        parser.add_argument("-a", "--archive", default=self.archive,
                            help=_("Replot transaction time data depending "
                                   "on the confidence levels configured by "
                                   "reading the archive file given."))
        parser.add_argument("-y", "--tpcbmax", default=self.tpcbmax, type=int,
                            help=_("Replot transaction rate and time "
                                   "diagram with user defined maximum at "
                                   "the y axis."))
        parser.add_argument("-b", "--database", default=self.job.db,
                            help=_("Relational database management system. "
                                   "Supported systems are: PostgreSQL and "
                                   "Firebird."))
        parser.add_argument("-t", "--transtype",
                            default=tpcbTransTable[self.job.transtype].symbol,
                            help=_("Type of transaction to be used for "
                                   "transaction rate measurement. Supported "
                                   "values are: TPCB_DEFAULT, TPCB_SELECT."))
        parser.add_argument("-o", "--output", default=self.output,
                            help=_("File name core (without extension) to "
                                   "save results."))
        parser.add_argument("-n", "--nodes", default=self.job.nodes, type=int,
                            help=_("Number of remote nodes performing "
                                   "transactions. The default value is 0, "
                                   "which means local processing only."))
        parser.add_argument("-j", "--threads", default=self.job.threads,
                            type=int,
                            help=_("Number of worker threads per node. "
                                   "A value of 0 determines that the main "
                                   "thread only is doing transactions."))
        parser.add_argument("-c", "--clients", default=self.job.clients,
                            type=int,
                            help=_("Number of clients/connections per "
                                   "thread."))
        parser.add_argument("-T", "--duration", default=self.job.duration,
                            type=float, help=_("Duration of one measurement "
                                               "cycle."))
        parser.add_argument("-i", "--no", default=self.no, type=int,
                            help=_("Number of measurement cycles per scale."))
        parser.add_argument("-s", "--scales", default=self.scales, type=int,
                            nargs="*", help=_("List of database scaling "
                                              "factors."))
        parser.add_argument("-d", "--daemon", default=self.daemon,
                            action="store_true",
                            help=_("Start application as daemon. As daemon or "
                                   "remote node it will wait for transaction "
                                   "requests from the master."))
        parser.add_argument("-k", "--killer", default=self.killer,
                            action="store_true",
                            help=_("Start application as node killer."))
        parser.add_argument("-p", "--prepare", default=self.job.prepare,
                            action="store_true",
                            help=_("Use prepare statements for transactions."))
        parser.add_argument("-r", "--remove", default=self.remove,
                            action="store_false",
                            help=_("Avoids to remove the database after "
                                   "measurements."))
        parser.add_argument("-x", "--reconnect",
                            default=self.job.reconnect,
                            action="store_true",
                            help=_("Connects and disconnects the database "
                                   "for each transaction."))
        parser.add_argument("-w", "--retry", default=self.job.retry,
                            action="store_true",
                            help=_("Retry failed transactions until they "
                                   "succeeded. The transaction time in this "
                                   "case is measured from the beginning of "
                                   "the first transaction attempt until the "
                                   "end of the transaction which succeeded."))
        parser.add_argument("-g", "--suppress", default=self.suppress,
                            action="store_true",
                            help=_("Suppress opening window that allows "
                                   "manual entering of test conditions."))
        parser.add_argument("-D", "--dsi", default=self.dsi,
                            action="store_true",
                            help=_("Run test with standard parameters "
                                   "to determine the DSI for the system."))

        args = parser.parse_args()
        if args.config and os.path.exists(args.config):
            # Read configuration from the configuration file that is given
            # by command line and ignore all other command line parameters.
            exec(open(args.config).read())
        else:
            db = args.database.strip().lower()
            if db == "fb" or db == "firebird":
                self.job.db = "FB"
            # elif db == "ora" or db == "oracle":
            #    self.job.db = "ORA"
            else:
                self.job.db = "PG"

            self.job.transtype = eval(args.transtype)
            self.job.nodes = args.nodes
            self.job.threads = args.threads
            self.job.clients = args.clients
            self.job.duration = args.duration
            self.job.prepare = args.prepare
            self.job.reconnect = args.reconnect
            self.job.retry = args.retry

            self.archive = args.archive
            self.tpcbmax = args.tpcbmax
            self.output = args.output
            self.no = args.no
            self.scales = args.scales
            self.daemon = args.daemon
            self.killer = args.killer
            self.remove = args.remove
            self.suppress = args.suppress
            self.dsi = args.dsi

        if args.dsi:
            # If DSI option is set, override standard parameters to
            # determine the DSI locally for the system under test.
            self.saveraw = False
            self.accurate = False
            self.daemon = False
            self.killer = False
            self.remove = True
            self.no = 10
            self.scales = [1, 2, 5, 10, 20, 50, 100, 200, 500,
                           1000, 2000, 5000, 10000]
            self.job.pgServer = "localhost"
            self.job.fbServer = "localhost"
            self.job.transtype = TPCB_DEFAULT
            self.job.nodes = 0
            self.job.threads = 16
            self.job.clients = 1
            self.job.duration = 60.0
            self.job.prepare = False
            self.job.reconnect = False
            self.job.retry = False

    def replace_localhost(self, server):
        """Replace database server.

        If remote nodes receive as database server the strings "localhost",
        "127.0.0.1" or ":1" it will be replaced by s new server string which
        contains the IP or host name of the master.

        Args:
            server: String containing the IP or host name of the database
                server.
        """
        keys = ("pgServer", "fbServer")
        for key in keys:
            value = getattr(self.job, key)
            if value in LOCALTUPLE:
                setattr(self.job, key, server)

    def guiconfig(self):
        """Manual configuration.

        Opens a window that demands to user to enter test conditions.
        the function is ignored if graphical interface is not available.

        Returns:
            A boolean that is set to False if the user presses the
            Cancel button to abort testing.
        """
        if self.suppress or self.dsi:
            return True

        try:
            win = tpcbWin(self.job.db, self.job.transtype, self.output,
                          self.job.nodes, self.job.threads, self.job.clients,
                          self.job.duration, self.no, self.scales,
                          self.job.prepare, self.remove, self.job.reconnect,
                          self.job.retry)
            retval = win.show()
            if retval == 2:
                self.job.db, self.job.transtype, self.output, self.job.nodes, \
                    self.job.threads, self.job.clients, self.job.duration, \
                    self.no, self.scales, self.job.prepare, self.remove, \
                    self.job.reconnect, self.job.retry = win.results()
                del win
                return True
            elif retval == 1:
                self.deferred.append(
                    _("Ignoring manual configuration. Proceed with defaults."))
                del win
                return True
            else:
                del win
                return False

        except NameError:
            # Missing graphical user interface results in failure when
            # tpcbGui.py is imported. This lead further to a NameError
            # exception while instantiating a tpcbWin().
            self.deferred.append(_("Graphical user interface not available."))
            return True

    def registernode(self, parameters, ip):
        """Register remote nodes.

        Registers information about remote nodes that responded to the UDP
        broadcasts. A list of remote nodes is built at the first usages of
        this function. If the list contains as many elements as nodes
        defined all further nodes that send responses are rejected.

        Args:
            parameters: The parameters part of a node identification
                message. It must contain Python code that create a new
                tpcbNodeInfo() class instance.
            ip: IP address of the remote node.

        Returns;
            A boolean that is set to False if a remote node is rejected,
        """
        nodeinfo = eval(parameters)

        self.lock.acquire()
        try:
            for n in self.nodeinfo:
                if n.host == nodeinfo.host:
                    self.logger.info(SELECTNODE, nodeinfo.host, ip)
                    return True

            if len(self.nodeinfo) < self.job.nodes:
                # Exchange IP addresses:
                # The server IP is delivered by remote nodes (nodeinfo.ip).
                # The remote node IP comes from the caller function.
                self.own.ip = nodeinfo.ip
                nodeinfo.ip = ip
                self.nodeinfo.append(nodeinfo)
                self.logger.info(SELECTNODE, nodeinfo.host, ip)
                return True
            else:
                return False
        finally:
            self.lock.release()

    def dbpar(self, title, value):
        """Database parameters.

        Take over database parameters for the data sheet.

        Args:
            title: The title of the database parameter.
            value: The database parameter value.
        """
        p = tpcbDbPar(title=title, value=value)
        self.dbenv.append(p)

    def initialise(self):
        """Determine dependent environment.

        Initialises the test environment after the database is selected
        and the database parameters were determined. It is needed for
        instance to determine the file system type.
        """
        assert self.dbpath, _("Variable self.dbpath is not set!")

        if self.dsi:
            self.blankinfo(_("=== Executing standard DSI sequence ==="))

        if os.name == "posix" or os.name == "nt":
            try:
                stdev = os.stat(self.dbpath).st_dev
                partitions = psutil.disk_partitions()
                for p in partitions:
                    if os.stat(p.mountpoint).st_dev == stdev:
                        self.dbdev = p.device
                        self.fstype = p.fstype
                        break
            except Exception as e:
                self.logger.warning(
                    _("Could not determine the file system type. "
                      "Error message '%s'."), str(e))
                self.fstype = UNKNOWN
        else:
            self.fstype = UNKNOWN

    def outline(self, line):
        """Single line logger.

        Pipes an output line to the data sheet that contains all test
        environment information and to the logging module.

        Args:
            line: String message to be logged.
        """
        self.sheet.write(line + "\n")
        self.logger.info(line)

    def outnode(self, node):
        """Output node information.

        Outputs all information of one node or master.

        Args:
            node: A tpcbNodeInfo() instance containing node information
                to be logged.
        """
        fmt = "{0:s}: {1:s}"
        self.sheet.write("\n")
        self.outline(fmt.format(_("Host name"), node.host))
        self.outline(fmt.format(_("IP address"), node.ip))
        self.outline(fmt.format(_("System type"), node.name))
        self.outline(fmt.format(_("System platform"), node.platform))
        self.outline(fmt.format(_("Architecture"), node.arch))
        self.outline(fmt.format(_("Python installation"), node.python))
        self.outline(fmt.format(_("PSYCOPG2 version"), node.psycopg2))
        self.outline(fmt.format(_("FDB version"), node.fdb))
        self.outline(fmt.format(_("Firebirdsql version"), node.firebirdsql))

        if node.cpu > 0:
            cpu = str(node.cpu)
        else:
            cpu = NOTACCENSSIBLE
        self.outline(fmt.format(_("Physical number of CPUs"), cpu))
        if node.core > 0:
            core = str(node.core)
        else:
            core = NOTACCENSSIBLE
        self.outline(fmt.format(_("Logical number of CPU cores"), core))

        self.outline("{0:s}: {1:.2f}".format(
            _("Available RAM [GB]"), node.ram))
        self.outline("{0:s}: {1:.2f}".format(
            _("Available SWAP [GB]"), node.swap))

    def datasheet(self):
        """Output data sheet.

        Outputs all test environment parameters to a data sheet and to
        the logging module. The data sheet is a file which is set as part
        of results.
        """
        filename = self.outpath("data")
        self.sheet = None
        try:
            self.sheet = open(filename, mode="wt", encoding="utf-8")

            fmt = "{0:s}: {1:s}"
            self.outline(_("MASTER"))
            if self.islocal:
                self.outline(_("RDBMS hosted locally."))
            else:
                self.outline(_("RDBMS hosted remotely."))
            if self.dsi:
                self.outline(_("Standard DSI test sequence requested."))
            self.outnode(self.own)

            if len(self.nodeinfo) > 0:
                self.sheet.write("\n")
                self.outline(_("NODES"))
                for n in self.nodeinfo:
                    self.outnode(n)

            self.sheet.write("\n")
            self.outline(_("TEST CONDITIONS"))
            self.sheet.write("\n")
            self.outline("{0:s}: {1:s}".format(
                TRANSTYPE,
                tpcbTransTable[self.job.transtype].descr))
            self.outline("{0:s}: {1:.1f}".format(
                _("Duration of measurement cycle [s]"), self.job.duration))
            self.outline("{0:s}: {1:d}".format(
                _("Number of measurements per scaling factor"), self.no))
            self.outline(fmt.format(
                _("List of scaling factors"), str(self.scales)))

            if self.job.nodes == 0:
                nodes = _("local processing")
            else:
                nodes = str(self.job.nodes)
            self.outline(fmt.format(NUMBERNODES, nodes))

            self.outline("{0:s}: {1:d}".format(
                NUMBERTHREADS, self.job.threads))
            self.outline("{0:s}: {1:d}".format(
                NUMBERCLIENTS, self.job.clients))
            self.outline("{0:s}: {1:s}".format(
                _("Use prepare for transactions"), str(self.job.prepare)))
            self.outline("{0:s}: {1:s}".format(
                _("Remove database after measurements"), str(self.remove)))
            self.outline("{0:s}: {1:s}".format(
                _("Reconnect database for each transaction"),
                str(self.job.reconnect)))
            self.outline("{0:s}: {1:s}".format(
                _("Retry failed transactions"), str(self.job.retry)))

            self.sheet.write("\n")
            self.outline(_("DATABASE CONFIGURATION"))
            self.sheet.write("\n")
            if self.islocal:
                self.outline(fmt.format(_("Database root"), self.dbdev))
                self.outline(fmt.format(_("File system type"), self.fstype))
            else:
                self.outline(fmt.format(_("Database root"), NOTACCENSSIBLE))
                self.outline(fmt.format(_("File system type"), NOTACCENSSIBLE))

            if len(self.dbenv) > 0:
                for p in self.dbenv:
                    self.outline("{0:s}: {1:s}".format(p.title, str(p.value)))

        except Exception as e:
            self.logger.error(
                _("Could not save data sheet to '%s'. Error message: '%s'"),
                filename, str(e))
        finally:
            if self.sheet:
                self.sheet.close()
            del self.sheet

    def makearchive(self):
        """Archives all results.

        All result files are combined to an archive.
        If temporary files are activated the archive is created
        temporary first and then copied to the working directory.
        """
        pwd = os.getcwd()
        try:
            os.chdir(self.tempdir.name)
            dbhost = self.dbhost.replace(".", "_")
            dirname = dbhost + "-" + self.job.db + "-" + \
                tpcbTransTable[self.job.transtype].ID + "-" + \
                time.strftime("%Y%m%d%H%M")
            os.mkdir(dirname)
            for file in os.listdir("."):
                if file.startswith(self.output):
                    if file.endswith(".log"):
                        # Windows OS can not move files that are still
                        # used like the log file. It must be copied
                        # instead of moving. Other results files are
                        # moved in order to save disk space.
                        shutil.copy(file, dirname)
                    else:
                        shutil.move(file, dirname)
            if self.archiveformat == "gztar":
                shutil.make_archive(dirname, "gztar", ".", dirname)
            else:
                shutil.make_archive(dirname, "zip", ".", dirname)
            if self.archiveformat == "gztar":
                shutil.copy(dirname + ".tar.gz", pwd)
            else:
                shutil.copy(dirname + ".zip", pwd)
        except Exception as e:
            self.logger.error(
                _("Could not create results archive! Error Message: '%s'"),
                str(e))
        finally:
            os.chdir(pwd)

    def readarchive(self):
        """Reads an archive.

        Reads and unpacks the archive file given for replotting some
        results.
        """
        # Analyse extension
        if self.archive.endswith(".tar.gz"):
            self.archiveformat = "gztar"
        elif self.archive.endswith(".zip"):
            self.archiveformat = "zip"
        else:
            # Determine missing extension by looking for
            # archive files being available
            if os.path.exists(self.archive + ".tar.gz"):
                self.archiveformat = "gztar"
                self.archive += ".tar.gz"
            elif os.path.exists(self.archive + ".zip"):
                self.archiveformat = "zip"
                self.archive += ".zip"
            else:
                raise Exception(
                    _("Could not find archive file '{0:s}'").format(
                        self.archive))

        # Doing all in temporary folder
        pwd = os.getcwd()
        try:
            # Copy archive to temporary folder
            shutil.copy(self.archive, self.tempdir.name)
            os.chdir(self.tempdir.name)

            # Split file name of the archive
            filename = os.path.basename(self.archive)
            if self.archiveformat == "gztar":
                dirname = filename[:-7]
                pattern = "^(.*)-(PG|FB)-(TPC-B|SELECT)-[0-9]{12}\.tar\.gz$"
            else:
                dirname = filename[:-4]
                pattern = "^(.*)-(PG|FB)-(TPC-B|SELECT)-[0-9]{12}\.zip$"

            # Extract host name, database id and transaction type
            # from archive file name
            dummy, \
                self.own.host, self.job.db, transtype, \
                dummy = re.split(pattern, filename)
            if transtype == "SELECT":
                self.job.transtype = TPCB_SELECT
            else:
                self.job.transtype = TPCB_DEFAULT
            # Unpack archive
            shutil.unpack_archive(filename)

            # Determine self.output (name of result files)
            # by inspecting names of all files
            self.output = ""
            for name in os.listdir(dirname):
                replace = name.replace("-", ".", 1)
                core, dummy = replace.split(".", 1)
                if self.output:
                    if self.output != core:
                        raise Exception(
                            _("Archive contains files of non standard naming "
                              "conventions."))
                else:
                    self.output = core

            # Determine self.job.nodes (number of remote nodes)
            # and self.job.threads (number of threads per node)
            # by inspecting the content of the data sheet
            filename = os.path.join(dirname, self.output + ".data")
            sheet = None
            try:
                sheet = open(filename, mode="r", encoding="utf-8")
                for line in sheet:
                    if line.startswith(NUMBERNODES):
                        try:
                            self.job.nodes = int(
                                line.replace(NUMBERNODES + ":", "").strip())
                        except:
                            self.job.nodes = 0
                    elif line.startswith(NUMBERTHREADS):
                        self.job.threads = int(
                            line.replace(NUMBERTHREADS + ":", "").strip())
            except Exception as e:
                self.logger.error(
                    _("Could not read data sheet! Error Message: '%s'"),
                    str(e))
            finally:
                if sheet:
                    sheet.close()

            # Determine scales from <output>.tpcb.<scale> files
            scales = []
            test = self.output + ".tpcb."
            for name in os.listdir(dirname):
                if name.startswith(test):
                    try:
                        scale = int(name.replace(test, ""))
                        scales.append(scale)
                    except:
                        pass
            scales.sort()
            self.scales = scales

            # Copy files to parent folder
            for name in os.listdir(dirname):
                if name.startswith(self.output):
                    source = os.path.join(dirname, name)
                    shutil.copy(source, ".")

        except Exception as e:
            self.logger.error(
                _("Could not read results archive '%s'! Error Message: '%s'"),
                self.archive, str(e))
        finally:
            os.chdir(pwd)


class tpcbNd():
    """Node data.

    This class manages TPC-B measurements for a single scale value
    collected on a single node.
    """
    def __init__(self, env):
        """Initialisation.

        Creates an empty list that contains the results.
        """
        self.env = env
        self.results = []

    def new(self):
        """Open a new results data transfer.

        This function initialises some variables to sum results.
        """
        self.done = 0
        self.errors = 0
        self.residence = 0.0
        self.elapsed = 0.0

    def add(self, results):
        """Add new results data of one thread.

        Results are added here.

        Args:
            results: tpcbResults() tuple containing the results
                collected by one thread.
        """
        self.ID = results.ID
        self.done += results.done
        self.errors += results.errors
        self.residence += results.residence
        self.elapsed += results.elapsed

    def finish(self):
        """Close results data transfer.

        Calculates averages and creates a whole node tpcbResults()
        tuple. This is appended to the results list of this node.
        """
        if self.residence > 0:
            tpsR = self.done / self.residence  # TODO
        else:
            tpsR = 0.0
        if self.elapsed > 0:
            tpsE = self.done / self.elapsed  # TODO
        else:
            tpsE = 0.0
        tpsD = self.done / self.env.job.duration
        s = tpcbResults(self.ID,
                        self.done, self.errors, self.residence, self.elapsed,
                        tpsR, tpsE, tpsD)
        self.results.append(s)

    def append(self, results):
        """Append results.

        This function appends results immediately to the list without
        averaging.
        """
        self.results.append(results)

    def statistics(self):
        """Statistical analyse.

        Statistical analysation of transaction time measurements of this
        scaling factor and node.
        """
        # Number of results
        no = len(self.results)
        assert no > 0, _("No results available to calculate statistics!")

        # Sum and averaging
        values = []
        self.done = 0
        self.errors = 0
        self.residence = 0.0
        self.elapsed = 0.0
        self.mean = 0.0
        for result in self.results:
            self.done += result.done
            self.errors += result.errors
            self.residence += result.residence
            self.elapsed += result.elapsed
            self.mean += result.tpsD
            values.append(result.tpsD)
        self.mean = self.mean / no
        # Standard deviation for tpsD
        self.stdw = 0.0
        if no > 1:
            for result in self.results:
                q = result.tpsD - self.mean
                self.stdw += q * q
            self.stdw = math.sqrt(self.stdw / (no - 1))
        # Median and extrem values for tpsD
        values.sort()
        n = no // 2
        if no % 2 == 0:
            self.median = (values[n-1] + values[n]) / 2
        else:
            self.median = values[n]
        self.minimum = values[0]
        self.maximum = values[-1]


class tpcbSingle():
    """Data acquisition.

    This class is responsible for acquisition of TPC-B measurements for
    a single scale value. It stores transaction rates, makes statistics and
    also acquires concurrent time measurements dynamically.
    """
    def __init__(self, env, scale):
        """Initialisation.

        Determines defaults and strings required for results presentation.

        Args:
            env: Test environment, as tpcbEnv().
            scale: Scaling factor of the database.
        """
        self.init = True
        self.stat = True
        self.finished = False
        self.scale = scale
        self.env = env
        self.keys = []
        self.values = []
        self.nd = {}

        # Transaction residence times buffer
        self.buf = tpcbBuf(BLOCKSIZE, env.confidences)
        self.checkbuf = False

        # Multiple used strings
        self.scale_str = SCALINGFACTOR
        self.dbsize_str = DATABASESIZE
        self.median_str = _("{0:s} median [tps]").format(
            tpcbTransTable[env.job.transtype].descr)
        self.mean_str = _("{0:s} average [tps]").format(
            tpcbTransTable[env.job.transtype].descr)
        self.stdw_str = _("{0:s} standard deviation [tps]").format(
            tpcbTransTable[env.job.transtype].descr)
        self.min_str = _("{0:s} minimum [tps]").format(
            tpcbTransTable[env.job.transtype].descr)
        self.max_str = _("{0:s} maximum [tps]").format(
            tpcbTransTable[env.job.transtype].descr)
        self.done_str = _("number of transactions done")
        self.errors_str = _("number of failed transactions")
        self.residence_str = _("total transaction residence time [s]")
        self.elapsed_str = _("effective duration measured [s]")
        self.multi_str = _("transaction rate multiplicator")
        self.conf_str = tpcbTransTable[env.job.transtype].descr + \
            _(" for {0:.1f}% confidence level [tps]")
        self.duration_str = _("duration [s]")

    def setvalues(self, **kwargs):
        """Enter result values.

        This method enters results of measurements by using keyword
        arguments.

        Args:
            **kwargs: Keyword arguments that contain results. The keys
                'results' and 'dbsize' are predefined and used for
                transaction measurement results as well as database size.
                Other keys are dynamically used for concurrent time
                measurements that must be defined by the method
                tpcbMultiple.settimedata()
        """
        if kwargs is None:
            return
        self.stat = True
        self.finished = False

        for key, value in kwargs.items():

            if key == "results":
                # First usage: Initialise dictionary
                if self.init:
                    full = tpcbNd(self.env)
                    self.nd["FULL"] = full
                    if self.env.job.nodes > 1:
                        for r in value:
                            if r.ID not in self.nd.keys():
                                nd = tpcbNd(self.env)
                                self.nd[r.ID] = nd
                    self.init = False

                # Distibute and store results
                if len(value) > 1:
                    for nd in self.nd.values():
                        nd.new()
                    for r in value:
                        # Register results for all
                        self.nd["FULL"].add(r)
                        # Register results per node if requested
                        if self.env.job.nodes > 1:
                            if r.ID in self.nd.keys():
                                self.nd[r.ID].add(r)
                            else:
                                # Results from unknown node
                                msg = _("Node ID '{0:s}' is not "
                                        "registered.").format(r.ID)
                                raise Exception(msg)
                    for nd in self.nd.values():
                        nd.finish()
                else:
                    # Only one result tuple is stored
                    for r in value:
                        self.nd["FULL"].append(r)

            elif key == "dbsize":
                self.dbsize = value

            else:
                if key not in self.keys:
                    self.keys.append(key)
                setattr(self, key, value)

    def done(self):
        """Mark as finished.

        This method marks the data record as finished. This means,
        all measurements for this special scaling factor are completed,
        all data are acquired.
        """
        self.finished = True

    def statistics(self):
        """Statistical analyse.

        Statistical analysation of transaction time measurements of this
        scaling factor.
        """
        # Do it for each node
        for nd in self.nd.values():
            nd.statistics()

        # Confidence intervals
        self.multi = self.nd["FULL"].residence \
            / self.env.no / self.env.job.duration
        self.confidences = self.buf.statistics(self.multi)
        # Flag: statistics done
        self.stat = False

    def csvhead(self, timedata):
        """Title for CSV.

        Args:
            timedata: A dictionary that contains the time data that
                are collected.

        Returns:
            A title line containing all column titles for *.csv export.
        """
        out = '"{0:s}";"{1:s}";"{2:s}";"{3:s}";"{4:s}";"{5:s}";' \
              '"{6:s}";"{7:s}";"{8:s}";"{9:s}";"{10:s}";"{11:s}"'.format(
                  self.scale_str, self.dbsize_str, self.median_str,
                  self.mean_str, self.stdw_str, self.min_str, self.max_str,
                  self.done_str, self.errors_str, self.residence_str,
                  self.elapsed_str, self.multi_str)

        fmt = ';"{0:s}"'.format(self.conf_str)
        for conf in self.buf.confidences:
            out += fmt.format(conf)

        for key in self.keys:
            out += ';"{0:s} {1:s}"'.format(
                timedata[key].text, self.duration_str)
        out += "\n"
        return out

    def csvline(self):
        """Data line for CSV.

        Returns:
            A line for *.csv export containing all results.
        """
        if self.stat:
            self.statistics()

        out = "{0:d};{1:.3f};{2:.2f};{3:.2f};{4:.2f};{5:.2f};" \
              "{6:.2f};{7:d};{8:d};{9:.2f};{10:.2f};{11:.4f}".format(
                  self.scale, self.dbsize, self.nd["FULL"].median,
                  self.nd["FULL"].mean, self.nd["FULL"].stdw,
                  self.nd["FULL"].minimum, self.nd["FULL"].maximum,
                  self.nd["FULL"].done, self.nd["FULL"].errors,
                  self.nd["FULL"].residence, self.nd["FULL"].elapsed,
                  self.multi)

        for conf in self.confidences:
            out += ";{0:.2f}".format(conf)

        for key in self.keys:
            value = getattr(self, key)
            out += ";{0:.2f}".format(value)
        return out

    def csvparse(self, line):
        """Parse CSV data line.

        At the moment only some variables are required for
        replotting rate and cumulative frequency diagrams.

        Args:
            line: CSV formatted line.
        """
        values = line.split(";")
        self.scale = int(values[0])
        self.dbsize = float(values[1])
        nd = tpcbNd(self.env)
        nd.median = float(values[2])
        nd.mean = float(values[3])
        nd.stdw = float(values[4])
        nd.minimum = float(values[5])
        nd.maximum = float(values[6])
        nd.done = int(values[7])
        nd.errors = int(values[8])
        nd.residence = float(values[9])
        nd.elapsed = float(values[10])
        self.nd["FULL"] = nd
        self.multi = float(values[11])

    def txthead(self, timedata):
        """Legend for results table.

        Args:
            timedata: A dictionary that contains the time data that
                are collected.

        Returns:
            A multi line legend that explains the content of all columns
            of the measurement results file. The order of columns is defined
            by the txtline() method.
        """
        out = "#  1: {0:s}\n#  2: {1:s}\n#  3: {2:s}\n" \
              "#  4: {3:s}\n#  5: {4:s}\n#  6: {5:s}\n" \
              "#  7: {6:s}\n#  8: {7:s}\n#  9: {8:s}\n" \
              "# 10: {9:s}\n# 11: {10:s}\n# 12: {11:s}\n".format(
                  self.scale_str, self.dbsize_str, self.median_str,
                  self.mean_str, self.stdw_str, self.min_str, self.max_str,
                  self.done_str, self.errors_str, self.residence_str,
                  self.elapsed_str, self.multi_str)

        n = COLUMN
        for conf in self.buf.confidences:
            fmt = "# {0:2d}: {1:s}\n".format(n, self.conf_str)
            out += fmt.format(conf)
            n += 1

        for key in self.keys:
            out += "# {0:2d}: {1:s} {2:s}\n".format(
                n, timedata[key].text, self.duration_str)
            n += 1

        out += "\n"
        out += "{0:>6s} {1:>8s} {2:>10s} {3:>10s} {4:>10s} {5:>10s} " \
               "{6:>10s} {7:>8s} {8:>8s} {9:>10s} {10:>10s} {11:>10s}".format(
                   "#1", "#2", "#3", "#4", "#5", "#6", "#7",
                   "#8", "#9", "#10", "#11", "#12")
        n = COLUMN
        for conf in self.buf.confidences:
            out += " {0:>10s}".format("#" + str(n))
            n += 1

        for key in self.keys:
            out += " {0:>10s}".format("#" + str(n))
            n += 1
        out += "\n\n"

        return out

    def txtline(self):
        """Results for all.

        The statistical analysis is done automatically and only if required.

        Returns:
            The results of measurements for one scaling factor in a well
            formatted single line string.
        """
        if self.stat:
            self.statistics()

        out = "{0:6d} {1:8.3f} {2:10.2f} {3:10.2f} {4:10.2f} {5:10.2f} " \
              "{6:10.2f} {7:8d} {8:8d} {9:10.2f} {10:10.2f} {11:10.4f}".format(
                  self.scale, self.dbsize, self.nd["FULL"].median,
                  self.nd["FULL"].mean, self.nd["FULL"].stdw,
                  self.nd["FULL"].minimum, self.nd["FULL"].maximum,
                  self.nd["FULL"].done, self.nd["FULL"].errors,
                  self.nd["FULL"].residence, self.nd["FULL"].elapsed,
                  self.multi)

        for conf in self.confidences:
            out += " {0:10.2f}".format(conf)

        for key in self.keys:
            value = getattr(self, key)
            out += " {0:10.2f}".format(value)
        return out

    def nodehead(self):
        """Legend for node results table.

        Returns:
            A multi line legend that explains the content of all columns
            of the measurement results file. The order of columns is defined
            by the nodeline() method.
        """
        out = "#  1: {0:s}\n#  2: {1:s}\n#  3: {2:s}\n" \
              "#  4: {3:s}\n#  5: {4:s}\n#  6: {5:s}\n" \
              "#  7: {6:s}\n#  8: {7:s}\n#  9: {8:s}\n" \
              "# 10: {9:s}\n# 11: {10:s}\n".format(
                  self.scale_str, self.dbsize_str, self.median_str,
                  self.mean_str, self.stdw_str, self.min_str, self.max_str,
                  self.done_str, self.errors_str, self.residence_str,
                  self.elapsed_str)

        out += "\n"
        out += "{0:>6s} {1:>8s} {2:>10s} {3:>10s} {4:>10s} {5:>10s} " \
               "{6:>10s} {7:>8s} {8:>8s} {9:>10s} {10:>10s}".format(
                   "#1", "#2", "#3", "#4", "#5", "#6", "#7",
                   "#8", "#9", "#10", "#11")

        out += "\n\n"
        return out

    def nodeline(self, node):
        """Results for nodes.

        The statistical analysis is done automatically and only if required.

        Args:
            node: The ID of the node, this ist the host name.

        Returns:
            The results of measurements for one scaling factor on one node
            in a well formatted single line string.
        """
        if self.stat:
            self.statistics()

        out = "{0:6d} {1:8.3f} {2:10.2f} {3:10.2f} {4:10.2f} {5:10.2f} " \
              "{6:10.2f} {7:8d} {8:8d} {9:10.2f} {10:10.2f}".format(
                  self.scale, self.dbsize, self.nd[node].median,
                  self.nd[node].mean, self.nd[node].stdw,
                  self.nd[node].minimum, self.nd[node].maximum,
                  self.nd[node].done, self.nd[node].errors,
                  self.nd[node].residence, self.nd[node].elapsed)

        return out

    def replothead(self):
        """Legend for confidence replotting.

        Returns:
            A multi line legend that explains the content of all columns
            of the confidence replotting file.
        """
        n = 2
        out = "#  1: {0:s}\n".format(self.scale_str)
        for conf in self.buf.confidences:
            fmt = "# {0:2d}: {1:s}\n".format(n, self.conf_str)
            out += fmt.format(conf)
            n += 1
        out += "\n"
        return out

    def replotline(self):
        """Confidence reults.

        Returns:
            A line containing scale and transaction times for confidence
            levels configured.
        """
        out = "{0:6d}".format(self.scale)
        for conf in self.confidences:
            out += " {0:10.2f}".format(conf)
        return out

    def savebuf(self):
        """Save residence times.

        Saves the residence time buffer to a raw file and a table for
        cumulative frequency diagram plotting.
        """
        if self.buf:
            # Save original data as raw file
            if self.env.saveraw:
                filename = self.env.outpath("raw." + str(self.scale) + ".gz")
                try:
                    self.buf.savetxt(filename)
                except Exception as e:
                    self.env.logger.error(ERRSAVERESULTS, filename, str(e))

            # Save table for cumulative frequency diagram
            # depending on the configuration option "accurate"
            a = self.buf.sort()
            size = len(a)
            fmt = "{0:e} {1:e}\n"
            filename = self.env.outpath("tpcb." + str(self.scale))
            cf = None
            try:
                cf = open(filename, mode="wt", encoding="utf-8")
                # Output time values in milliseconds
                if self.env.accurate:
                    n = size
                    for rt in a:
                        cf.write(fmt.format(1000.0 * rt, n / size))
                        n -= 1
                else:
                    x_index = self.buf.redindex()
                    for x in x_index:
                        cf.write(fmt.format(1000.0 * a[x], (size - x) / size))
            except Exception as e:
                self.env.logger.error(ERRSAVERESULTS, filename, str(e))
            finally:
                if cf:
                    cf.close()

    def loadbuf(self):
        """Load residence times.

        Loads the residence time buffer data from a raw file.
        """
        filename = self.env.inpath("raw." + str(self.scale) + ".gz")
        try:
            self.buf.loadtxt(filename)
        except Exception as e:
            self.env.logger.error(ERRREADRESULTS, filename, str(e))

    def __repr__(self):
        """Results for logging.

        The statistical analysis is done automatically and only if required.

        Returns:
            An unformatted representation of measurements for one scaling
            factor for logging purposes.
        """
        if self.stat:
            self.statistics()

        out = "{0:d} {1:.3f} {2:.2f} {3:.2f} {4:.2f} {5:.2f} " \
              "{6:.2f} {7:d} {8:d} {9:.2f} {10:.2f} {11:.4f}".format(
                  self.scale, self.dbsize, self.nd["FULL"].median,
                  self.nd["FULL"].mean, self.nd["FULL"].stdw,
                  self.nd["FULL"].minimum, self.nd["FULL"].maximum,
                  self.nd["FULL"].done, self.nd["FULL"].errors,
                  self.nd["FULL"].residence, self.nd["FULL"].elapsed,
                  self.multi)

        for conf in self.confidences:
            out += " {0:.2f}".format(conf)

        for key in self.keys:
            value = getattr(self, key)
            out += " {0:.2f}".format(value)

        if self.finished:
            out += " => ok"
        else:
            out += " => not finished"
        return out


class tpcbMultiple():
    """Measurements table.

    This class collents transaction rates, concurrent times and so on
    depending on the scale facor. Imagine this as a table where rows
    represent different scale factors and columns the different issues
    defined in the class tpcbSingle().
    """
    def __init__(self, env):
        """Initialisation.

        Initialises variables for graphical representation of test results
        and the records that gain single scale data.

        Args:
            env: The test environment, a tpcbEnv() instance.
        """
        self.stat = True
        self.available = False
        self.frequencies = False
        self.env = env
        self.ymax = {}
        self.nymax = 0.0
        self.cmax = 0.0
        self.tmax = 0.0
        self.tmin = sys.float_info.max
        self.maxsize = 0.0
        self.DSI = 0.0
        self.minsize = sys.maxsize
        self.singles = []
        for scale in self.env.scales:
            m = tpcbSingle(self.env, scale)
            self.singles.append(m)
        self.markers = []
        self.timedata = {}
        # See tpcbSingle.txtline():
        # The first column at which tps at definite confidence levels
        # and concurrent time measurements can be found in the result's
        # table is stored here as variable self.n_times.
        self.n_times = COLUMN + len(env.confidences)

    def clean(self):
        """Clean data buffers.

        This functions releases memory allocated by the residence time
        buffers.
        """
        for m in self.singles:
            m.buf.clean()

    def setmarker(self, gb, text):
        """Registration of marks.

        Registration of marks for memory sizes like RAM, SWAP or
        database configurations and their labels for the diagram.

        Args:
            gb: Memory site value in GB.
            text: Label for the diagram to describe the mark.
        """
        self.markers.append(tpcbMarker(gb=gb, text=text))

    def settimedata(self, key, text):
        """Registration of time measurements.

        Registration of concurrent time measurements like duration
        of database initialisation, index generation, drop database
        containing the column number (which is assigned automatically)
        and a label for column description.

        Args:
            key: Keyword for a specific time.
            text: Descriptive string of that time.
        """
        self.timedata[key] = tpcbTimeData(nr=self.n_times, text=text)
        self.n_times += 1

    def initialise(self):
        """Diagram initialisation.

        Preparation of diagram output. It determines axes scales from the
        measurements.
        """
        i = 0
        nr = 0
        s = []
        y = []

        self.stat = False
        self.fx0 = 1000000.0
        self.fx1 = 0.0

        for m in self.singles:
            if m.buf:
                self.frequencies = True
                nr += m.buf.nr
                m.checkbuf = True
                mintime, maxtime = m.buf.minmax()
                if mintime < self.fx0:
                    self.fx0 = mintime
                if maxtime > self.fx1:
                    self.fx1 = maxtime

            if m.finished:
                if i > 0:
                    # Set flag 'available' only if at least 2 scale values
                    # are completed. This flag controls the generation of
                    # graphical output.
                    self.available = True
                else:
                    # Initialise ymax dictionary
                    for key in m.nd.keys():
                        self.ymax[key] = 0.0
                i += 1

                # Fill arrays for DSI calculation
                s.append(math.log10(m.scale))
                y.append(math.log10(m.nd["FULL"].median))

                for key in self.timedata.keys():
                    value = getattr(m, key)
                    if value > self.tmax:
                        self.tmax = value
                    if value < self.tmin:
                        self.tmin = value

                if m.dbsize > self.maxsize:
                    self.maxsize = m.dbsize
                if m.dbsize < self.minsize:
                    self.minsize = m.dbsize

                for key in m.nd.keys():
                    if m.nd[key].maximum > self.ymax[key]:
                        self.ymax[key] = m.nd[key].maximum
                    if m.nd[key].mean + m.nd[key].stdw > self.ymax[key]:
                        self.ymax[key] = m.nd[key].mean + m.nd[key].stdw

                if self.env.job.nodes > 1:
                    for key in m.nd.keys():
                        if key != "FULL":
                            if self.ymax[key] > self.nymax:
                                self.nymax = self.ymax[key]

                if m.confidences is not None:
                    for conf in m.confidences:
                        if conf > self.cmax:
                            self.cmax = conf

        # DSI calculation
        if self.available:
            dsi = 0.0
            for i in range(len(s) - 1):
                dsi += (y[i+1] + y[i]) * (s[i+1] - s[i]) / 2.0
            self.DSI = dsi / 2.0

    def results(self):
        """Save results to disk.

        Output:
            * Logging DSI
            *.txt: formatted text file including legend suitable for
                diagram generation by means of gnuplot
            *.csv: unformatted semicolon separated for imports in office
                calc programs
            *.cyc: formatted table containing tpsD results of each cycle
            *.tpcb.<scale>: unformatted cumulative frequency data of
                transaction times
        """
        if self.stat:
            self.initialise()
        if not self.available:
            return

        if self.env.dsi:
            self.env.logger.info(_("Standard DSI sequence: DSI=%.1f (%f)."),
                                 round(self.DSI, 1), self.DSI)
        else:
            self.env.logger.info(_("DSI=%.1f (%f)."),
                                 round(self.DSI, 1), self.DSI)

        filename = self.env.outpath("txt")
        txt = None
        try:
            txt = open(filename, mode="wt", encoding="utf-8")
            # Legend for columns
            txt.write(self.singles[0].txthead(self.timedata))
            # Write measurements
            for m in self.singles:
                if m.finished:
                    txt.write(m.txtline() + "\n")
        except Exception as e:
            self.env.logger.error(ERRSAVERESULTS, filename, str(e))
        finally:
            if txt:
                txt.close()

        filename = self.env.outpath("csv")
        csv = None
        try:
            csv = open(filename, mode="wt", encoding="utf-8")
            # Column titles
            csv.write(self.singles[0].csvhead(self.timedata))
            # Write measurements
            for m in self.singles:
                if m.finished:
                    csv.write(m.csvline() + "\n")
        except Exception as e:
            self.env.logger.error(ERRSAVERESULTS, filename, str(e))
        finally:
            if csv:
                csv.close()

        if self.env.job.nodes > 1:
            # Plot rates per node if at least 2 nodes included,
            # for each node a separate file.
            for nodeinfo in self.env.nodeinfo:
                filename = self.env.outpath("node." + nodeinfo.host)
                node = None
                try:
                    node = open(filename, mode="wt", encoding="utf-8")
                    # Legend for columns
                    node.write(self.singles[0].nodehead())
                    # Write node specific measurements
                    for m in self.singles:
                        if m.finished:
                            node.write(m.nodeline(nodeinfo.host) + "\n")
                except Exception as e:
                    self.env.logger.error(ERRSAVERESULTS, filename, str(e))
                finally:
                    if node:
                        node.close()

        if self.env.no > 1:
            # Plot rates depending on cycle number
            # only if demanded number of cycles is more than 1.
            filename = self.env.outpath("cyc")
            cyc = None
            try:
                cyc = open(filename, mode="wt", encoding="utf-8")
                # Title containing scales
                out = "#  "
                for m in self.singles:
                    if m.finished:
                        out += " {0:10d}".format(m.scale)
                cyc.write(out + "\n\n")
                # Write measurements depending on cycle
                for i in range(self.env.no):
                    cycle = i + 1
                    out = "{0:3d}".format(cycle)
                    for m in self.singles:
                        if m.finished:
                            out += " {0:10.2f}".format(
                                m.nd["FULL"].results[i].tpsD)
                    cyc.write(out + "\n")
            except Exception as e:
                self.env.logger.error(ERRSAVERESULTS, filename, str(e))
            finally:
                if cyc:
                    cyc.close()

        for m in self.singles:
            m.savebuf()

    def read(self):
        """Read results from disk.

        This method reads files containing unformatted and raw residence
        time data and recalculate the transaction rate for the confidence
        intervals.
        """
        filename = self.env.inpath("csv")
        csv = None
        try:
            csv = open(filename, mode="r", encoding="utf-8")
            # Waste column titles
            csv.readline()
            # Read measurements
            for m in self.singles:
                m.csvparse(csv.readline())
                m.done()
        except Exception as e:
            self.env.logger.error(ERRREADRESULTS, filename, str(e))
        finally:
            if csv:
                csv.close()

        for m in self.singles:
            # Load raw residence time data
            m.loadbuf()
            # Recalculate transaction rates for new confidence intervals
            m.confidences = m.buf.statistics(m.multi)

    def winFindGnuplot(self):
        """Find gnuplot.

        This method tries to find wgnuplot.exe on Windows OS
        if not configured.

        Returns:
            The absolute path of the wgnuplot.exe executable.
        """
        gnuplot = self.env.gnuplot
        if os.path.exists(gnuplot):
            return gnuplot
        for root, dirs, files in os.walk("C:\\"):
            for f in files:
                if f == "wgnuplot.exe":
                    gnuplot = os.path.join(root, f)
                    # Store file search result for further usages
                    self.env.gnuplot = gnuplot
                    return gnuplot
        return None

    def gnuplot(self, filename):
        """Execute gnuplot.

        This method tries to execute gnuplot, on demand in a temporary
        folder.

        Args:
            filename: Name of the script containing gnuplot instructions.
        """
        pwd = os.getcwd()
        try:
            os.chdir(self.env.tempdir.name)
            if os.name == "nt":
                gnuplot = self.winFindGnuplot()
                if gnuplot:
                    os.system("{0:s} {1:s}".format(gnuplot, filename))
                else:
                    self.env.logger.error(
                        _("Could not find gnuplot executable!"))
            else:
                os.system("gnuplot " + filename)
        except Exception as e:
            self.env.logger.error(
                _("Could not execute gnuplot! Error message: '%s'"), str(e))
        finally:
            os.chdir(pwd)

    def plotmarks(self, idx, ymax, file):
        """Plot marks.

        Plots memory size marks in the diagram.

        Args:
            idx: Index,
                0 for left hand orientation (linear scale),
                1 for right hand orientation (logarithmic scale).
            ymax: Maximum y axis value.
            file: File handle to write the gnuplot instructions to.
        """
        arrow = ("set arrow from second {0:f}, first {1:f} to "
                 "second {2:f}, first {3:f} head\n",
                 "set arrow from second {0:f}, second {1:f} to "
                 "second {2:f}, second {3:f} head\n")
        label_t = ("set label rotate \"{0:s} \" at second {1:f}, "
                   "first {2:f} right font \"Times-Roman,8\"\n",
                   "set label rotate \"{0:s} \" at second {1:f}, "
                   "second {2:f} right font \"Times-Roman,8\"\n")
        label_b = ("set label rotate \" {0:s}\" at second {1:f}, "
                   "first {2:f} left font \"Times-Roman,8\"\n",
                   "set label rotate \" {0:s}\" at second {1:f}, "
                   "second {2:f} left font \"Times-Roman,8\"\n")

        ydiff = ymax / 20
        yup = ymax - ydiff

        for marker in self.markers:
            # Only mark within the axes range
            if not self.minsize < marker.gb < self.maxsize:
                continue

            # Select mark at upper or lower position
            # depending on the transaction rate average at the database
            # size nearby
            bottom = True
            for m in self.singles:
                if m.dbsize > marker.gb:
                    if idx == 1:
                        test = math.log(m.confidences[0], 10) \
                            < math.log(ymax / 2, 10)
                    else:
                        test = m.nd["FULL"].mean < ymax / 2
                    if test:
                        bottom = False
                    break

            if bottom:
                file.write(arrow[idx].format(marker.gb, ydiff, marker.gb, 0))
                file.write(label_b[idx].format(marker.text, marker.gb, ydiff))
            else:
                file.write(arrow[idx].format(marker.gb, yup, marker.gb, ymax))
                file.write(label_t[idx].format(marker.text, marker.gb, yup))

    def plotrates(self):
        """Transaction rates and time diagram.

        Saves a gnuplot script that creates a diagram from the measurement
        results. This diagram contains the averaged TPC-B transaction rates
        as well as the concurrent time measurement results.
        """
        if self.stat:
            self.initialise()
        if not self.available:
            return

        # Transaction rate diagram
        filename = self.env.outpath("rate.plot")
        file = None
        try:
            file = open(filename, mode="wt", encoding="utf-8")

            file.write("set encoding utf8\n")
            file.write("set term pdf color\n")
            file.write("set key below\n")
            file.write("set key font \",8\"\n")
            file.write("set style fill transparent solid 0.1 noborder\n")
            file.write("set output \"{0:s}-rate.pdf\"\n".format(
                self.env.output))
            file.write("set grid xtics ytics\n")

            # Definition of axes
            file.write("set xtics 1,10 nomirror\n")
            file.write("set mxtics 10\n")
            file.write("set logscale x\n")
            file.write("set xlabel \"{0:s}\"\n".format(SCALINGFACTOR))
            file.write("set xrange[{0:d}:{1:d}]\n".format(self.env.scales[0],
                                                          self.env.scales[-1]))

            file.write("set x2tics 0.1,10\n")
            file.write("set mx2tics 10\n")
            file.write("set logscale x2\n")
            file.write("set x2label \"{0:s}\"\n".format(DATABASESIZE))
            file.write("set x2range[{0:f}:{1:f}]\n".format(self.minsize,
                                                           self.maxsize))

            if self.env.tpcbmax:
                ymax = self.env.tpcbmax
            else:
                ymax = 1.01 * self.ymax["FULL"]
            file.write("set ytics nomirror\n")
            file.write("set ylabel \"{0:s}\"\n".format(_("{0:s} [tps]").format(
                tpcbTransTable[self.env.job.transtype].descr)))
            file.write("set yrange[0:{0:f}]\n".format(ymax))

            file.write("set y2tics 0.01,10\n")
            file.write("set my2tics 10\n")
            file.write("set logscale y2\n")
            file.write("set y2label \"{0:s}\"\n".format(
                _("duration [s]\\n (see legend below)")))
            if self.tmin <= 0:
                file.write("set y2range[*:{0:f}]\n".format(self.tmax))
            else:
                file.write("set y2range[{0:f}:{1:f}]\n".format(self.tmin,
                                                               self.tmax))

            # Set marks
            self.plotmarks(0, ymax, file)

            # Plot transaction results
            file.write("plot \\\n")
            # Min/Max with hatching
            file.write("  \"" + self.env.output + ".txt\" u 1:6:7 axis x1y1 "
                       "t '' w filledcurves lt 2, \\\n")
            file.write("  \"" + self.env.output + ".txt\" u 1:6 axis x1y1 "
                       "t '' w lines lt 2, \\\n")
            file.write("  \"" + self.env.output + ".txt\" u 1:7 axis x1y1 "
                       "t '' w lines lt 2, \\\n")
            # Line for Median Value
            file.write("  \"" + self.env.output + ".txt\" u 1:3 axis x1y1 "
                       "t '' w lines lt 1, \\\n")
            # Errorbars for Mean Value
            file.write("  \"" + self.env.output + ".txt\" u 1:4:5 axis x1y1 "
                       "t '{0:s}' w errorbars pt 1 lt 1".format(
                           tpcbTransTable[self.env.job.transtype].descr))

            # Plot concurrent time measurements
            colorwheel = ("678D00", "000000", "FF00FF",
                          "00FFFF", "FF9C00", "0000FF")

            for key in self.timedata.keys():
                # Avoid manually excluded
                if key in self.env.exclctimes:
                    continue
                # Avoid zero values on the log scale
                ignore = False
                for m in self.singles:
                    if m.finished:
                        t = getattr(m, key)
                        if t <= 0.0:
                            ignore = True
                            break
                if ignore:
                    continue
                # Integration of one line for time measurements
                file.write(",\\\n  \"" + self.env.output + ".txt\" u 1:{0:d} "
                           "axis x1y2 t '{1:s}' w lines lt rgb "
                           "\"#{2:s}\"".format(
                               self.timedata[key].nr, self.timedata[key].text,
                               colorwheel[self.timedata[key].nr % 6]))

            file.write("\n")
        except Exception as e:
            self.env.logger.error(ERRGNUPLOT, filename, str(e))
        finally:
            if file:
                file.close()

        # Invoke gnuplot
        self.gnuplot(self.env.output + ".rate.plot")

    def replotrates(self):
        """Transaction rates diagram replotted.

        Saves a gnuplot script that creates a diagram from the measurement
        results. This diagram contains the averaged TPC-B transaction rates
        only. It is used for replotting.
        """
        if self.stat:
            self.initialise()
        if not self.available:
            return

        # Transaction rate diagram
        filename = self.env.outpath("rate.replot.plot")
        file = None
        try:
            file = open(filename, mode="wt", encoding="utf-8")

            file.write("set encoding utf8\n")
            file.write("set term pdf color\n")
            file.write("set nokey\n")
            file.write("set style fill transparent solid 0.1 noborder\n")
            file.write("set output \"{0:s}-rate.replot.pdf\"\n".format(
                self.env.output))
            file.write("set grid xtics ytics\n")

            # Definition of axes
            file.write("set xtics 1,10 nomirror\n")
            file.write("set mxtics 10\n")
            file.write("set logscale x\n")
            file.write("set xlabel \"{0:s}\"\n".format(SCALINGFACTOR))
            file.write("set xrange[{0:d}:{1:d}]\n".format(self.env.scales[0],
                                                          self.env.scales[-1]))

            file.write("set x2tics 0.1,10\n")
            file.write("set mx2tics 10\n")
            file.write("set logscale x2\n")
            file.write("set x2label \"{0:s}\"\n".format(DATABASESIZE))
            file.write("set x2range[{0:f}:{1:f}]\n".format(self.minsize,
                                                           self.maxsize))

            if self.env.tpcbmax:
                ymax = self.env.tpcbmax
            else:
                ymax = 1.01 * self.ymax["FULL"]
            file.write("set ytics mirror\n")
            file.write("set ylabel \"{0:s}\"\n".format(_("{0:s} [tps]").format(
                tpcbTransTable[self.env.job.transtype].descr)))
            file.write("set yrange[0:{0:f}]\n".format(ymax))

            # Use <output>.rate.plot to read the markers
            pattern = "^set label rotate \"(.*)\" at second ([0-9\.]+), first"
            source = self.env.outpath("rate.plot")
            src = None
            try:
                src = open(source, mode="r", encoding="utf-8")
                for line in src:
                    if re.match(pattern, line):
                        dummy, text, gb, dummy = re.split(pattern, line)
                        self.markers.append(tpcbMarker(gb=float(gb), text=text))
            except Exception as e:
                self.env.logger.error(ERRREADPLOT, source, str(e))
            finally:
                if src:
                    src.close()

            # Set marks
            self.plotmarks(0, ymax, file)

            # Plot transaction results
            file.write("plot \\\n")
            # Min/Max with hatching
            file.write("  \"" + self.env.output + ".txt\" u 1:6:7 axis x1y1 "
                       "w filledcurves lt 2, \\\n")
            file.write("  \"" + self.env.output + ".txt\" u 1:6 axis x1y1 "
                       "w lines lt 2, \\\n")
            file.write("  \"" + self.env.output + ".txt\" u 1:7 axis x1y1 "
                       "w lines lt 2, \\\n")
            # Line for Median Value
            file.write("  \"" + self.env.output + ".txt\" u 1:3 axis x1y1 "
                       "w lines lt 1, \\\n")
            # Errorbars for Mean Value
            file.write("  \"" + self.env.output + ".txt\" u 1:4:5 axis x1y1 "
                       "w errorbars pt 1 lt 1")

            file.write("\n")
        except Exception as e:
            self.env.logger.error(ERRGNUPLOT, filename, str(e))
            raise e
        finally:
            if file:
                file.close()

        # Invoke gnuplot
        self.gnuplot(self.env.output + ".rate.replot.plot")

    def plotnodes(self):
        """Transaction rates per node diagram.

        Saves a gnuplot script that creates a diagram from the measurement
        results per node. This diagram contains the averaged transaction
        rates only.
        """
        # Do not plot if only one node
        if self.env.job.nodes <= 1:
            return

        if self.stat:
            self.initialise()
        if not self.available:
            return

        # Transaction rate diagram per node
        filename = self.env.outpath("node.plot")
        file = None
        try:
            file = open(filename, mode="wt", encoding="utf-8")

            file.write("set encoding utf8\n")
            file.write("set term pdf color\n")
            file.write("set key outside right\n")
            file.write("set key font \",8\"\n")
            file.write("set output \"{0:s}-node.pdf\"\n".format(
                self.env.output))
            file.write("set grid xtics ytics\n")

            # Definition of axes
            file.write("set xtics 1,10 nomirror\n")
            file.write("set mxtics 10\n")
            file.write("set logscale x\n")
            file.write("set xlabel \"{0:s}\"\n".format(SCALINGFACTOR))
            file.write("set xrange[{0:d}:{1:d}]\n".format(self.env.scales[0],
                                                          self.env.scales[-1]))

            file.write("set x2tics 0.1,10\n")
            file.write("set mx2tics 10\n")
            file.write("set logscale x2\n")
            file.write("set x2label \"{0:s}\"\n".format(DATABASESIZE))
            file.write("set x2range[{0:f}:{1:f}]\n".format(self.minsize,
                                                           self.maxsize))

            if self.env.tpcbmax:
                ymax = self.env.tpcbmax
            else:
                ymax = 1.01 * self.nymax
            # file.write("set logscale y\n")
            file.write("set ylabel \"{0:s}\"\n".format(_("{0:s} [tps]").format(
                tpcbTransTable[self.env.job.transtype].descr)))
            file.write("set yrange[0:{0:f}]\n".format(ymax))

            # Plot transaction results per node
            lt = 1
            first = True
            file.write("plot \\\n")
            for nodeinfo in self.env.nodeinfo:
                if first:
                    first = False
                else:
                    file.write(", \\\n")
                name = self.env.output + ".node." + nodeinfo.host
                file.write("  \"{0:s}\" u 1:4:6:7 axis x1y1 t '{1:s}' "
                           "w errorbars lt {2:d}, \\\n".format(name,
                                                               nodeinfo.host,
                                                               lt))
                file.write("  \"{0:s}\" u 1:3 axis x1y1 t '' w lines "
                           "lt {1:d}".format(name, lt))
                lt += 1

        except Exception as e:
            self.env.logger.error(ERRGNUPLOT, filename, str(e))
        finally:
            if file:
                file.close()

        # Invoke gnuplot
        self.gnuplot(self.env.output + ".node.plot")

    def plotcycles(self):
        """Transaction rates on cycles diagram.

        Saves a gnuplot script that creates a diagram containing transaction
        rates of different scales depending on the cycle number. This ia a
        limited representation how transaction rates evolve in time after
        database creation.
        """
        if self.env.no < 2:
            return
        if self.stat:
            self.initialise()
        if not self.available:
            return

        # Transaction rate on cycles diagram
        filename = self.env.outpath("cycle.plot")
        file = None
        try:
            file = open(filename, mode="wt", encoding="utf-8")

            file.write("set encoding utf8\n")
            file.write("set term pdf color\n")
            file.write("set key outside right\n")
            file.write("set key font \",8\"\n")
            file.write("set output \"{0:s}-cycle.pdf\"\n".format(
                self.env.output))
            file.write("set grid xtics ytics\n")

            # Definition of axes
            file.write("set logscale y\n")
            file.write("set mytics 10\n")
            file.write("set ylabel \"{0:s}\"\n".format(_("{0:s} [tps]").format(
                tpcbTransTable[self.env.job.transtype].descr)))
            if self.env.no <= 10:
                file.write("set xtics 1\n")
            file.write("set xlabel \"{0:s}\"\n".format(_("number of cycle")))
            file.write("set xrange[0:{0:d}]\n".format(self.env.no + 1))

            # Plotting data
            n = 2
            file.write("plot \\\n")
            for m in self.singles:
                if m.finished:
                    if n > 2:
                        file.write(", \\\n")
                    file.write("  \"" + self.env.output + ".cyc\" u 1:{0:d} "
                               "t \"{1:d}\" w lines".format(n, m.scale))
                    n += 1

        except Exception as e:
            self.env.logger.error(ERRGNUPLOT, filename, str(e))
        finally:
            if file:
                file.close()

        # Invoke gnuplot
        self.gnuplot(self.env.output + ".cycle.plot")

    def plotconfidences(self):
        """Transaction rates for confidence levels diagram.

        Saves a gnuplot script that creates a diagram from the measurement
        results. This diagram contains the TPC-B transaction rates for
        different configured confidence levels.
        """
        if self.stat:
            self.initialise()
        if not self.available:
            return

        # Transaction rate on scales diagram
        filename = self.env.outpath("conf.plot")
        file = None
        try:
            file = open(filename, mode="wt", encoding="utf-8")

            file.write("set encoding utf8\n")
            file.write("set term pdf color\n")
            file.write("set key outside right\n")
            file.write("set key font \",8\"\n")
            file.write("set output \"{0:s}-conf.pdf\"\n".format(
                self.env.output))
            file.write("set grid xtics ytics\n")

            # Definition of axes
            file.write("set xtics 1,10 nomirror\n")
            file.write("set mxtics 10\n")
            file.write("set logscale x\n")
            file.write("set xlabel \"{0:s}\"\n".format(SCALINGFACTOR))
            file.write("set xrange[{0:d}:{1:d}]\n".format(self.env.scales[0],
                                                          self.env.scales[-1]))

            file.write("set x2tics 0.1,10\n")
            file.write("set mx2tics 10\n")
            file.write("set logscale x2\n")
            file.write("set x2label \"{0:s}\"\n".format(DATABASESIZE))
            file.write("set x2range[{0:f}:{1:f}]\n".format(self.minsize,
                                                           self.maxsize))

            file.write("set logscale y\n")
            file.write("set mytics 10\n")
            file.write("set ytics nomirror\n")
            file.write("set ylabel \"{0:s}\"\n".format(_("{0:s} [tps]").format(
                tpcbTransTable[self.env.job.transtype].descr)))

            cmax = 1.01 * self.cmax
            file.write("set y2range[0:{0:f}]\n".format(cmax))

            # Set marks
            self.plotmarks(1, cmax, file)

            # Plot transaction results
            file.write("plot \\\n")
            n = COLUMN
            for conf in self.env.confidences:
                if n > COLUMN:
                    file.write(", \\\n")
                file.write("  \"" + self.env.output + ".txt\" u 1:{0:d} axis "
                           "x1y1 t '{1:.1f}%' w lines".format(n, conf))
                n += 1
        except Exception as e:
            self.env.logger.error(ERRGNUPLOT, filename, str(e))
        finally:
            if file:
                file.close()

        # Invoke gnuplot
        self.gnuplot(self.env.output + ".conf.plot")

    def replotconfidences(self):
        """Update transaction rates for confidence levels diagram.

        Saves a new file containing transaction rates for confidence levels
        only and creates a new diagram from this.
        """
        filename = self.env.outpath("replot")
        replot = None
        try:
            replot = open(filename, mode="wt", encoding="utf-8")
            # Table legend
            replot.write(self.singles[0].replothead())
            # Write results of recalculated confidence levels
            for m in self.singles:
                replot.write(m.replotline() + "\n")
        except Exception as e:
            self.env.logger.error(ERRSAVERESULTS, filename, str(e))
        finally:
            if replot:
                replot.close()

        # Use <output>.conf.plot as template gnuplot script
        source = self.env.outpath("conf.plot")
        file = None
        try:
            file = open(source, mode="r", encoding="utf-8")
            content = file.readlines()
        except Exception as e:
            self.env.logger.error(ERRREADPLOT, source, str(e))
        finally:
            if file:
                file.close()

        # Create a new diagram from the template by exchanging some
        # lines
        destination = self.env.outpath("conf.replot.plot")
        file = None
        try:
            file = open(destination, mode="wt", encoding="utf-8")
            for line in content:
                if line.startswith("set output"):
                    # Exchange *.pdf file name
                    file.write("set output \"{0:s}-conf.replot.pdf\"\n".format(
                        self.env.output))
                elif not line.startswith("plot"):
                    # Copy all lines from template
                    file.write(line)
                else:
                    # Exchange all lines after the "plot" command
                    n = 2
                    file.write(line)
                    for conf in self.env.confidences:
                        if n > 2:
                            file.write(", \\\n")
                        file.write("  \"" + self.env.output + ".replot\" "
                                   "u 1:{0:d} axis x1y1 t '{1:.1f}%' "
                                   "w lines".format(n, conf))
                        n += 1
                    break

        except Exception as e:
            self.env.logger.error(ERRGNUPLOT, destination, str(e))
        finally:
            if file:
                file.close()

        # Invoke gnuplot
        self.gnuplot(self.env.output + ".conf.replot.plot")

    def plotfrequencies(self):
        """Cumulative frequency diagram.

        Saves a gnuplot script that displays cumulative frequencies
        expressed as confidence levels of transaction residence times
        for all scales.
        """
        if self.stat:
            self.initialise()
        if not self.frequencies:
            return

        filename = self.env.outpath("freq.plot")
        file = None
        try:
            file = open(filename, mode="wt", encoding="utf-8")

            file.write("set encoding utf8\n")
            file.write("set term pdf color\n")
            file.write("set key outside right\n")
            file.write("set key font \",8\"\n")
            file.write("set output \"{0:s}-freq.pdf\"\n".format(
                self.env.output))
            file.write("set logscale x\n")
            file.write("set mxtics 10\n")
            file.write("set xlabel \"{0:s}\"\n".format(
                _("transaction residence time [ms]")))
            file.write("set logscale y\n")
            file.write("set mytics 10\n")
            file.write("set yrange[*:1.5]\n")
            file.write("set ylabel \"{0:s}\"\n".format(
                _("confidence level [%]")))
            tics = "'0' 1"
            for i in range(6):
                y = math.pow(10.0, -(i+1))
                fmt = ", '{0:." + str(max(i-1, 0)) + "f}' {1:f}"
                tics += fmt.format(100.0 * (1.0 - y), y)
            file.write("set ytics ({0:s})\n".format(tics))
            file.write("set grid xtics ytics\n")

            f = lambda x: math.pow(10, math.floor(math.log(1000 * x, 10)))
            x0 = f(self.fx0)
            x1 = f(self.fx1)
            arrow = "set arrow from {0:f}, {1:f} to {2:f}, {3:f} head\n"
            label = "set label \" {0:.1f}%\" at {1:f}, {2:f} left font " \
                    "\"Times-Roman,8\"\n"
            for conf in self.env.confidences:
                y = 1.0 - (conf / 100.0)
                file.write(arrow.format(x1, y, x0, y))
                file.write(label.format(conf, x1, y))

            file.write("plot \\\n")
            end = False
            for m in self.singles:
                # Check residence time buffer indirectly using 'checkbuf'
                # because the buffer is cleared here to free memory for the
                # gnuplot run.
                if m.checkbuf:
                    if end:
                        file.write(", \\\n")
                    else:
                        end = True
                    name = self.env.output + ".tpcb.{0:d}".format(m.scale)
                    file.write("  \"" + name + "\" u 1:2 t \"{0:d}\" "
                               "w lines".format(m.scale))
            file.write("\n")
        except Exception as e:
            self.env.logger.error(ERRGNUPLOT, filename, str(e))
        finally:
            if file:
                file.close()

        # Invoke gnuplot
        self.gnuplot(self.env.output + ".freq.plot")

    def presentresults(self):
        """Present all results.

        This method calls all routines that make data representation
        as tables and graphics.
        """
        # Save results including diagram creation with gnuplot
        self.results()
        self.plotrates()
        self.plotnodes()
        self.plotcycles()
        self.plotconfidences()
        # Note:
        # Before plotting the cumulative frequencies the allocated
        # memory for residence times is released. All other data are
        # small compared to residence times.
        self.clean()
        self.plotfrequencies()
