# -*- coding: utf-8 -*-
#
# dsbench/tpcbFirebirdSql.py - Firebird driver based on firebirdsql
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

from tpcbCore import *
from tpcbDbFb import *


# i18n preparations.
def _(s):
    return s


class tpcbDbFirebirdSql(tpcbDbFb):
    """Database controller for Firebird.

    This controller overrides tpcbDb() class method for accessing a
    Firebird database based on the 'firebirdsql' driver.
    """
    def __init__(self, env, multiple):
        """Initialisation.

        Args:
            env: The test environment, a tpcbEnv() instance.
            multiple: The measurements table, a tpcbMultiple() instance.
        """
        if not FIREBIRDSQL:
            raise Exception(_("Python module 'firebirdsql' not installed."))

        tpcbDbFb.__init__(self, env, multiple)

    def _connect(self):
        """Connect to the database.

        This is an override method for the Firebird database.

        Returns:
            A connection and a cursor to the database.
        """
        con = None
        cur = None
        if not self.env.job.reconnect:
            self.env.logger.debug(
                _("Connect to Firebird database '%s' on '%s'."),
                self.dbname, self.server)
        con = firebirdsql.connect(dsn=self.dsn, user=self.user_name,
                                  password=self.password, charset="utf8",
                                  page_size=16384)
        # con.default_tpb = firebirdsql.ISOLATION_LEVEL_REPEATABLE_READ
        cur = con.cursor()
        return con, cur

    def _disconnect(self, con, cur):
        """Disconnect from database.

        This is an override method for the Firebird database.

        Args:
            con: Connection to the database
            cur: Cursor from the database connection
        """
        if not self.env.job.reconnect:
            self.env.logger.debug(
                _("Disconnect from Firebird database '%s' on '%s'."),
                self.dbname, self.server)
        con.commit()
        cur.close()
        con.close()

    def _create(self, scale):
        """Create a database.

        This is an override method for the Firebird database.
        An extisting database having the same name will be deleted!

        Args:
            scale: Scaling factor of the database.

        Returns:
            A tuple containing durations in seconds for filling the
            database, creating the primary and foreign keys.
        """
        self.env.logger.info(
            _("Recreating Firebird database '%s' on '%s' with scale=%d."),
            self.dbname, self.server, scale)

        # Drop old database
        con = None
        try:
            con = firebirdsql.connect(dsn=self.dsn, user=self.user_name,
                                      password=self.password, charset="utf8",
                                      page_size=16384)
            con.drop_database()
            con.close()
            self.env.logger.debug(_("Database '%s' dropped."), self.dbname)
        except:
            pass

        # Create new database
        con = None
        con = firebirdsql.create_database(dsn=self.dsn, user=self.user_name,
                                          password=self.password,
                                          charset="utf8", page_size=16384)
        cur = con.cursor()
        self.env.logger.debug(_("Database '%s' created."), self.dbname)

        # Tables
        self.fileexec(_("Table 'tpcb_branch' created."),
                      "tpcb_branch.sql", con, cur)
        self.fileexec(_("Table 'tpcb_teller' created."),
                      "tpcb_teller.sql", con, cur)
        self.fileexec(_("Table 'tpcb_account' created."),
                      "tpcb_account.sql", con, cur)
        self.fileexec(_("Table 'tpcb_history' created."),
                      "tpcb_history.sql", con, cur)

        # Functions
        self.fileexec(_("Function 'func_init' created."),
                      "func_init.sql", con, cur)
        self.fileexec(_("Function 'func_work' created."),
                      "func_work.sql", con, cur)
        self.fileexec(_("Function 'func_work_select' created."),
                      "func_work_select.sql", con, cur)

        # Initialisation
        tInit = self.timeexec(
            _("Initializing TPC-B with scale={0:d}.").format(scale),
            "execute procedure tpcb_func_p_init(1, 10, 100000, {0:d})".format(
                scale),
            con, cur)

        # Create Index
        tPk = 0.0
        tFk = 0.0
        # Branch
        tPk += self.timeexec(
            _("Creating primary key on 'tpcb_branch'."),
            "alter table tpcb_branch add constraint tpcb_branch_pk "
            "primary key (branch_id)",
            con, cur)

        # Teller
        tPk += self.timeexec(
            _("Creating primary key on 'tpcb_teller'."),
            "alter table tpcb_teller add constraint tpcb_teller_pk "
            "primary key (teller_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_branch' on 'tpcb_teller'."),
            "alter table tpcb_teller add constraint tpcb_teller_fk "
            "foreign key (branch_id) references tpcb_branch(branch_id)",
            con, cur)

        # Account
        tPk += self.timeexec(
            _("Creating primary key on 'tpcb_account'."),
            "alter table tpcb_account add constraint tpcb_account_pk "
            "primary key (account_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_branch' on 'tpcb_account'."),
            "alter table tpcb_account add constraint tpcb_account_fk "
            "foreign key (branch_id) references tpcb_branch(branch_id)",
            con, cur)

        # History
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_branch' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_1 "
            "foreign key (branch_id) references tpcb_branch(branch_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_teller' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_2 "
            "foreign key (teller_id) references tpcb_teller(teller_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_account' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_3 "
            "foreign key (account_id) references tpcb_account(account_id)",
            con, cur)

        self.env.logger.debug("t(INIT)=%.2f, t(PK)=%.2f, t(FK)=%.2f",
                              tInit, tPk, tFk)

        self._disconnect(con, cur)
        return tInit, tPk, tFk

    def _drop(self):
        """Drop the database.

        This is an override method for the Firebird database.

        Returns:
            Time needed to drop the database in seconds.
        """
        self.env.logger.info(_("Drop Firebird database '%s'."), self.dbname)

        con, cur = self._connect()
        start = timeit.default_timer()
        con.drop_database()
        end = timeit.default_timer()
        self.env.logger.debug(_("Database '%s' dropped."), self.dbname)
        cur.close()
        con.close()

        tDrop = end - start
        self.env.logger.debug("t(DROP)=%.2f", tDrop)
        return tDrop

    def _clean(self):
        """Clean the database between measurement cycles.

        This is an override method for the Firebird database.
        Cleans the database between several transaction rate measurements.
        Firebird tries this to achive the same results as the _clean()
        method in the PostgreSQL module. It deletes the whole content of the
        'tpcb_history' table.

        Returns:
            Time needed to clean the database in seconds.
        """
        con, cur = self._connect()

        tClean = 0.0

        # Drop foreign keys
        tClean += self.timeexec(
            _("Dropping foreign key to 'tpcb_account' on 'tpcb_history'."),
            "alter table tpcb_history drop constraint tpcb_history_fk_3",
            con, cur)
        tClean += self.timeexec(
            _("Dropping foreign key to 'tpcb_teller' on 'tpcb_history'."),
            "alter table tpcb_history drop constraint tpcb_history_fk_2",
            con, cur)
        tClean += self.timeexec(
            _("Dropping foreign key to 'tpcb_branch' on 'tpcb_history'."),
            "alter table tpcb_history drop constraint tpcb_history_fk_1",
            con, cur)
        # Drop functions
        tClean += self.timeexec(
            _("Dropping function 'func_work_select'."),
            "drop procedure tpcb_func_p_work_select",
            con, cur)
        tClean += self.timeexec(
            _("Dropping function 'func_work'."),
            "drop procedure tpcb_func_p_work",
            con, cur)
        # Drop table
        tClean += self.timeexec(
            _("dropping table tpcb_history"),
            "drop table tpcb_history",
            con, cur)

        # Recreate table
        tClean += self.fileexec(_("Table 'tpcb_history' created."),
                                "tpcb_history.sql", con, cur)
        # Reinstall functions
        tClean += self.fileexec(_("Function 'func_work' created."),
                                "func_work.sql", con, cur)
        tClean += self.fileexec(_("Function 'func_work_select' created."),
                                "func_work_select.sql", con, cur)
        # Recreate foreign keys
        tClean += self.timeexec(
            _("Creating foreign key to 'tpcb_branch' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_1 "
            "foreign key (branch_id) references tpcb_branch(branch_id)",
            con, cur)
        tClean += self.timeexec(
            _("Creating foreign key to 'tpcb_teller' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_2 "
            "foreign key (teller_id) references tpcb_teller(teller_id)",
            con, cur)
        tClean += self.timeexec(
            _("Creating foreign key to 'tpcb_account' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_3 "
            "foreign key (account_id) references tpcb_account(account_id)",
            con, cur)

        self._disconnect(con, cur)
        self.env.logger.debug("t(CLEAN)=%.2f", tClean)
        return tClean

    def _size(self):
        """Determine the physical size of the database in GB.

        This is an override method for the Firebird database.

        Returns:
            The size of the database in GB as a float.
        """
        con, cur = self._connect()
        res = con.db_info([firebirdsql.isc_info_page_size,
                           firebirdsql.isc_info_allocation])
        pagesAllocated = res[firebirdsql.isc_info_allocation]
        pageSize = res[firebirdsql.isc_info_page_size]
        self._disconnect(con, cur)
        return pageSize * pagesAllocated / GB

    def _execute(self, sql, con, cur):
        """Executes SQL commands.

        This is an override method for the Firebird database.

        Args:
            sql: SQL statements.
            con: Connection to the database.
            cur: Cursor on the connection.
        """
        cur.execute(sql)
        con.commit()

    def prepare(self, cur):
        """Prepare.

        Makes a preparation for the TPC-B transaction statement.

        Args:
            cur: Cursor on the connection.

        Returns:
            A prepared statement for execution.
        """
        sql = "execute procedure {0:s}(?, ?, ?, ?)".format(
            tpcbTransTable[self.env.job.transtype].func)
        return cur.prep(sql)

    def command(self, branch_id, teller_id, account_id, delta):
        """Transaction statement.

        Creates the SQL command for one TPC-B transaction.
        Delivers different output depending on prepare is used.

        Args:
            branch_id: ID of branch.
            teller_id: ID of teller.
            account_id: ID of account.
            delta: Amount.

        Returns:
            An SQL statement as string.
        """
        sql = "execute procedure {0:s}({1:d}, {2:d}, {3:d}, {4:d})".format(
            tpcbTransTable[self.env.job.transtype].func,
            branch_id, teller_id, account_id, delta)
        # sql = "select o_account_balance from {0:s}(" \
        #      "{1:d}, {2:d}, {3:d}, {4:d})".format(
        #          tpcbTransTable[self.env.job.transtype].func,
        #          branch_id, teller_id, account_id, delta)
        return sql

    def transaction(self, sql, con, cur):
        """Execute transaction.

        Executes a single transaction including return value pickup.
        Firebird requires an explicit commit.

        Args:
            sql: SQL statements.
            con: Connection to the database.
            cur: Cursor on the connection.
        """
        cur.execute(sql)
        x = cur.fetchone()[0]
        con.commit()

    def prepared(self, prep, con, cur,
                 branch_id, teller_id, account_id, delta):
        """Execute prepared.

        Executes a prepared transaction including return value pickup.

        Args:
            prep: Prepared statement.
            branch_id: ID of branch.
            teller_id: ID of teller.
            account_id: ID of account.
            delta: Amount.
        """
        cur.execute(prep, (branch_id, teller_id, account_id, delta))
        x = cur.fetchone()[0]
        con.commit()

    def chkerror(self, e):
        """Error check.

        Checks if the exception is caused by a temporary database error
        that is normal operation.

        Args:
            e: Exception to be checked.

        Returns:
            A boolean that is False if the error is normal operation.
        """
        if not isinstance(e, firebirdsql.fbcore.DatabaseError):
            return True
        else:
            return False

    def _loop_single(self, evkill):
        """Transaction loop with single connection.

        This is an override method for the Firebird database.
        Each transaction opens and closes the database connection.

        Args:
            evkill: A thread event to signalise kill.

        Returns:
            A tuple containing the number of successful transactions done,
            the number of rejected transactions due to collisions, the total
            transaction residence time and effective duration elapsed during
            transactions.
        """
        finish = False
        residence = 0.0
        elapsed = 0.0
        done = 0
        errors = 0
        t0 = timeit.default_timer()

        # Run transaction's cycle
        while True:
            # Random numbers for transaction input values
            branch_id, teller_id, account_id, delta = self.rand.next()

            # Open connection and prepare transaction
            t1 = timeit.default_timer()
            con, cur = self._connect()
            if self.env.job.prepare:
                prep = self.prepare(cur)
            else:
                sql = self.command(branch_id, teller_id, account_id, delta)

            # Do transaction
            while True:
                ok = True
                try:
                    if self.env.job.prepare:
                        self.prepared(prep, con, cur,
                                      branch_id, teller_id, account_id,
                                      delta)
                    else:
                        self.transaction(sql, con, cur)
                    self._disconnect(con, cur)
                except Exception as e:
                    # Catch errors while transactions
                    if self.chkerror(e):
                        raise e
                    ok = False
                finally:
                    # Get current time and check duration
                    t2 = timeit.default_timer()
                    if t2 - t0 > self.env.job.duration:
                        # Mark as duration finished, waste transaction
                        # and break the inner retry loop
                        finish = True
                        break
                    else:
                        # Register effective duration
                        elapsed = t2 - t0
                        if ok:
                            # Transaction succeeded
                            # Register transaction residence time
                            diff = t2 - t1
                            self.single.buf.set(diff)
                            # Sum total residence time
                            residence += diff
                            done += 1
                            break
                        else:
                            # Transaction failed
                            errors += 1
                            # Retrying transaction until success
                            if not self.env.job.retry:
                                break

            # Abort loop due to end of duration
            if finish:
                break

            # Abort loop if thread receives kill request
            if evkill and evkill.isSet():
                break

        return done, errors, residence, elapsed

    def _loop_multi(self, evkill, cons, curs):
        """Transaction loop with multiple connections.

        This is an override method for the Firebird database.
        Connections are opened and closed only once for the whole
        measurement cycle.

        Args:
            evkill: A thread event to signalise kill.
            cons: A list of connections to the database.
            curs: A list of cursors on the connections.

        Returns:
            A tuple containing the number of successful transactions done,
            the number of rejected transactions due to collisions, the total
            transaction residence time and effective duration elapsed during
            transactions.
        """
        # Prepare
        if self.env.job.prepare:
            preps = []
            for cur in curs:
                prep = self.prepare(cur)
                preps.append(prep)

        # Initialise cycle
        c = 0
        cmax = len(curs)
        finish = False
        residence = 0.0
        elapsed = 0.0
        done = 0
        errors = 0
        t0 = timeit.default_timer()
        t1 = t0

        # Run transaction's cycle
        while True:
            # Use clients via "round robin"
            c = (c + 1) % cmax
            # Random numbers for transaction input values
            branch_id, teller_id, account_id, delta = self.rand.next()
            if not self.env.job.prepare:
                sql = self.command(branch_id, teller_id, account_id, delta)

            # Do transaction
            t1 = timeit.default_timer()
            while True:
                ok = True
                try:
                    if self.env.job.prepare:
                        self.prepared(preps[c], cons[c], curs[c],
                                      branch_id, teller_id, account_id,
                                      delta)
                    else:
                        self.transaction(sql, cons[c], curs[c])
                except Exception as e:
                    # Catch errors while transactions
                    if self.chkerror(e):
                        raise e
                    ok = False
                finally:
                    # Get current time and check duration
                    t2 = timeit.default_timer()
                    if t2 - t0 > self.env.job.duration:
                        # Mark as duration finished, waste transaction
                        # and break the inner retry loop
                        finish = True
                        break
                    else:
                        # Register effective duration
                        elapsed = t2 - t0
                        if ok:
                            # Transaction succeeded
                            # Register transaction residence time
                            diff = t2 - t1
                            self.single.buf.set(diff)
                            # Sum total residence time
                            residence += diff
                            done += 1
                            break
                        else:
                            # Transaction failed
                            errors += 1
                            # Retrying transaction until success
                            if not self.env.job.retry:
                                break

            # Abort loop due to end of duration
            if finish:
                break

            # Abort loop if thread receives kill request
            if evkill and evkill.isSet():
                break

        return done, errors, residence, elapsed

    def _parameters(self):
        """Collect database parameters.

        This is an override method for the Firebird database.
        Some Firebird parameters are collected from a database that
        is created temporary. Other parameters are parsed from the database
        configuration file.
        """
        # Create a test database with scale = 1
        self._create(1)

        # Connect to that database and query parameters
        con, cur = self._connect()
        res = con.db_info([firebirdsql.isc_info_page_size])
        pagesize = res[firebirdsql.isc_info_page_size]
        self._disconnect(con, cur)

        # Default parameters
        isolation = "repeatable read"

        # Parse parameters from configuration file
        databaseAccess, defaultDbCachePages, \
            fileSystemCacheThreshold = self.readconfig(pagesize)

        # Apply database parameters
        if databaseAccess:
            self.env.dbpath = databaseAccess
            self.env.dbpar("DatabaseAccess", databaseAccess)
        else:
            self.env.dbpath = self.dbname

        self.env.dbpar("Firebird Python driver", "firebirdsql")
        self.env.dbpar("Transaction isolation", isolation)

        if fileSystemCacheThreshold:
            self.env.dbpar("FileSystemCacheThreshold [GB]",
                           fileSystemCacheThreshold)
            self.multiple.setmarker(fileSystemCacheThreshold,
                                    "FileSystemCacheThreshold")
        if defaultDbCachePages:
            self.env.dbpar("DefaultDbCachePages [GB]", defaultDbCachePages)
            self.multiple.setmarker(defaultDbCachePages, "DefaultDbCachePages")
